import logging
import re
import shutil
import subprocess
import sys
import time
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Tuple,
)
import uuid

from safety_schemas.models.events import EventType
from safety_schemas.models.events.types import ToolType
from safety_schemas.models.events.payloads import (
    FirewallConfiguredPayload,
    FirewallDisabledPayload,
    PackageInstalledPayload,
    PackageUninstalledPayload,
    PackageUpdatedPayload,
    CommandExecutedPayload,
    ToolCommandExecutedPayload,
    CommandErrorPayload,
    AliasConfig,
    IndexConfig,
    ToolStatus,
    CommandParam,
    ProcessStatus,
    FirewallHeartbeatPayload,
)

from ..event_bus import EventBus

from .creation import (
    create_event,
)
from .data import (
    clean_parameter,
    get_command_path,
    get_root_context,
    scrub_sensitive_value,
    translate_param_source,
)
from .conditions import conditional_emitter, should_emit_firewall_heartbeat


if TYPE_CHECKING:
    from safety.models import SafetyCLI, ToolResult
    from safety.cli_util import CustomContext

logger = logging.getLogger(__name__)


@conditional_emitter(conditions=[should_emit_firewall_heartbeat])
def emit_firewall_heartbeat(
    event_bus: "EventBus", ctx: Optional["CustomContext"], *, tools: List[ToolStatus]
):
    payload = FirewallHeartbeatPayload(tools=tools)
    event = create_event(payload=payload, event_type=EventType.FIREWALL_HEARTBEAT)

    event_bus.emit(event)


@conditional_emitter
def emit_firewall_disabled(
    event_bus: "EventBus",
    ctx: Optional["CustomContext"] = None,
    *,
    reason: Optional[str],
):
    payload = FirewallDisabledPayload(reason=reason)
    event = create_event(payload=payload, event_type=EventType.FIREWALL_DISABLED)

    event_bus.emit(event)


@conditional_emitter
def emit_firewall_configured(
    event_bus: "EventBus",
    ctx: Optional["CustomContext"] = None,
    *,
    alias_config: AliasConfig,
    index_config: IndexConfig,
):
    PIP_TOOL = "pip"
    command_path = shutil.which(PIP_TOOL)
    reachable = False
    version = "unknown"

    if command_path:
        args = [command_path, "--version"]
        result = subprocess.run(args, capture_output=True, text=True)

        if result.returncode == 0:
            output = result.stdout
            reachable = True

            # Extract version
            version_match = re.search(r"pip (\d+\.\d+(?:\.\d+)?)", output)
            if version_match:
                version = version_match.group(1)
    else:
        command_path = PIP_TOOL

    tool = ToolStatus(
        type=ToolType.PIP,
        command_path=command_path,
        version=version,
        reachable=reachable,
        alias_config=alias_config,
        index_config=index_config,
    )

    payload = FirewallConfiguredPayload(tools=[tool])

    event = create_event(payload=payload, event_type=EventType.FIREWALL_CONFIGURED)

    event_bus.emit(event)


@conditional_emitter
def emit_diff_operations(
    event_bus: "EventBus",
    ctx: "CustomContext",
    *,
    added: Dict[str, str],
    removed: Dict[str, str],
    updated: Dict[str, Tuple[str, str]],
    by_tool: ToolType,
):
    obj: "SafetyCLI" = ctx.obj
    correlation_id = obj.correlation_id

    if (added or removed or updated) and not correlation_id:
        correlation_id = obj.correlation_id = str(uuid.uuid4())

    def emit_package_event(event_bus, correlation_id, payload, event_type):
        event = create_event(
            payload=payload,
            event_type=event_type,
            correlation_id=correlation_id,
        )
        event_bus.emit(event)

    for package_name, version in added.items():
        emit_package_event(
            event_bus,
            correlation_id,
            PackageInstalledPayload(
                package_name=package_name, version=version, tool=by_tool
            ),
            EventType.PACKAGE_INSTALLED,
        )

    for package_name, version in removed.items():
        emit_package_event(
            event_bus,
            correlation_id,
            PackageUninstalledPayload(
                package_name=package_name, version=version, tool=by_tool
            ),
            EventType.PACKAGE_UNINSTALLED,
        )

    for package_name, (previous_version, current_version) in updated.items():
        emit_package_event(
            event_bus,
            correlation_id,
            PackageUpdatedPayload(
                package_name=package_name,
                previous_version=previous_version,
                current_version=current_version,
                tool=by_tool,
            ),
            EventType.PACKAGE_UPDATED,
        )


@conditional_emitter
def emit_tool_command_executed(
    event_bus: "EventBus", ctx: "CustomContext", *, tool: ToolType, result: "ToolResult"
) -> None:
    correlation_id = ctx.obj.correlation_id

    if not correlation_id:
        correlation_id = ctx.obj.correlation_id = str(uuid.uuid4())

    process = result.process

    payload = ToolCommandExecutedPayload(
        tool=tool,
        raw_command=[clean_parameter("", arg) for arg in process.args],
        duration_ms=result.duration_ms,
        status=ProcessStatus(
            stdout=process.stdout, stderr=process.stderr, return_code=process.returncode
        ),
    )

    # Scrub after binary coercion to str
    if payload.status.stdout:
        payload.status.stdout = scrub_sensitive_value(payload.status.stdout)
    if payload.status.stderr:
        payload.status.stderr = scrub_sensitive_value(payload.status.stderr)

    event = create_event(
        correlation_id=correlation_id,
        payload=payload,
        event_type=EventType.TOOL_COMMAND_EXECUTED,
    )

    event_bus.emit(event)


@conditional_emitter
def emit_command_executed(
    event_bus: "EventBus", ctx: "CustomContext", *, returned_code: int
) -> None:
    params: List[CommandParam] = []

    root_context = get_root_context(ctx)

    if root_context and hasattr(root_context, "started_at") and root_context.started_at:
        duration_ms = int((time.monotonic() - root_context.started_at) * 1000)
    else:
        duration_ms = 1

    command_name = ctx.command.name
    raw_command = [clean_parameter("", arg) for arg in sys.argv]

    params: List[CommandParam] = []

    for idx, param in enumerate(ctx.command.params):
        param_name = param.name
        param_value = ctx.params.get(param_name)

        # Scrub the parameter value if sensitive
        scrubbed_value = clean_parameter(param_name, param_value)

        # Determine parameter source using Click's API
        click_source = ctx.get_parameter_source(param_name)
        source = translate_param_source(click_source)

        display_name = param_name if param_name else None

        params.append(
            CommandParam(
                position=idx, name=display_name, value=scrubbed_value, source=source
            )
        )

    payload = CommandExecutedPayload(
        command_name=command_name,
        command_path=get_command_path(ctx),
        raw_command=raw_command,
        parameters=params,
        duration_ms=duration_ms,
        status=ProcessStatus(
            return_code=returned_code,
        ),
    )

    event = create_event(
        correlation_id=ctx.obj.correlation_id,
        payload=payload,
        event_type=EventType.COMMAND_EXECUTED,
    )

    try:
        if future := event_bus.emit(event):
            future.result(timeout=0.5)
    except Exception:
        logger.error("Emit Failed %s (%s)", event.type, event.id)


@conditional_emitter
def emit_command_error(
    event_bus: "EventBus",
    ctx: "CustomContext",
    *,
    message: str,
    traceback: Optional[str] = None,
) -> None:
    """
    Emit a CommandErrorEvent with sensitive data scrubbed.
    """
    # Get command name from context if available
    command_name = getattr(ctx, "command", None)
    if command_name and command_name.name:
        command_name = command_name.name

    scrub_traceback = None
    if traceback:
        scrub_traceback = scrub_sensitive_value(traceback)

    command_path = get_command_path(ctx)
    raw_command = [scrub_sensitive_value(arg) for arg in sys.argv]

    payload = CommandErrorPayload(
        command_name=command_name,
        raw_command=raw_command,
        command_path=command_path,
        error_message=scrub_sensitive_value(message),
        stacktrace=scrub_traceback,
    )

    event = create_event(
        payload=payload,
        event_type=EventType.COMMAND_ERROR,
    )

    event_bus.emit(event)



from sys import platform
import unittest
from unittest.mock import patch, mock_open
from pathlib import Path
import tempfile
import shutil

from datetime import datetime, timezone

import pytest

from safety.tool.interceptors.unix import UnixAliasInterceptor


@pytest.mark.unix_only
@pytest.mark.skipif(platform not in ["linux", "linux2", "darwin"], 
                    reason="Unix-specific tests")
class TestUnixAliasInterceptor(unittest.TestCase):

    def setUp(self):        
        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.temp_dir)

    @patch('safety.tool.interceptors.unix.Path.home')
    @patch('safety.tool.interceptors.base.datetime')
    @patch('safety.tool.interceptors.base.get_version')    
    def test_interceptors_all_tools(self, mock_version,
                                            mock_datetime,
                                            mock_home):

        mock_home.return_value = Path(self.temp_dir)
        mock_version.return_value = "1.0.0"
        mock_now = datetime(2024, 1, 1, tzinfo=timezone.utc)
        mock_datetime.now.return_value = mock_now

        safety_config_user_dir = Path(self.temp_dir) / '.safety'

        with patch('safety.tool.interceptors.unix.USER_CONFIG_DIR', 
                   safety_config_user_dir):            

          interceptor = UnixAliasInterceptor()
          result = interceptor.install_interceptors()

          self.assertTrue(result)
          
          profile_path = Path(self.temp_dir) / '.profile'
          safety_profile_path = Path(self.temp_dir) / '.safety' / '.safety_profile'
          
          self.assertTrue(profile_path.exists())
          self.assertTrue(safety_profile_path.exists())
          
          # test the content of the generated files
          expected_profile_content = (
              "# >>> Safety >>>\n"
              f'[ -f "{safety_profile_path}" ] && . "{safety_profile_path}"\n'
              "# <<< Safety <<<\n"
          )
          
          expected_safety_profile_content = (
              "# >>> Safety >>>\n"
              "# DO NOT EDIT THIS FILE DIRECTLY\n"
              f"# Last updated at: {mock_now.isoformat()}\n"
              "# Updated by: safety v1.0.0\n"
              'alias pip="safety pip"\n'
              'alias pip3="safety pip"\n'
              "# <<< Safety <<<\n"
          )
          
          self.assertEqual(profile_path.read_text(), expected_profile_content)
          self.assertEqual(safety_profile_path.read_text(),
                          expected_safety_profile_content)
          
          # Let's test remove_interceptors
          result = interceptor.remove_interceptors()
          
          self.assertTrue(result)          
          self.assertTrue(profile_path.exists())
          self.assertEqual(profile_path.read_text(), "")

          self.assertFalse(safety_profile_path.exists())          

    def test_install_interceptors_nonexistent_tool(self):        
        interceptor = UnixAliasInterceptor()
        result = interceptor.install_interceptors(['nonexistent'])
        self.assertFalse(result)

    def test_uninstall_interceptors_all_tools(self):
        interceptor = UnixAliasInterceptor()
        result = interceptor.install_interceptors()
        self.assertTrue(result)



import configparser

from typing import Any, Dict, Optional, Tuple, Union

from authlib.oidc.core import CodeIDToken
from authlib.jose import jwt
from authlib.jose.errors import ExpiredTokenError

from safety.auth.models import Organization
from safety.auth.constants import CLI_AUTH_LOGOUT, CLI_CALLBACK, AUTH_CONFIG_USER, CLI_AUTH
from safety.constants import CONFIG
from safety_schemas.models import Stage
from safety.util import get_proxy_dict


def get_authorization_data(client, code_verifier: str,
                           organization: Optional[Organization] = None,
                           sign_up: bool = False, ensure_auth: bool = False, headless: bool = False) -> Tuple[str, str]:
    """
    Generate the authorization URL for the authentication process.

    Args:
        client: The authentication client.
        code_verifier (str): The code verifier for the PKCE flow.
        organization (Optional[Organization]): The organization to authenticate with.
        sign_up (bool): Whether the URL is for sign-up.
        ensure_auth (bool): Whether to ensure authentication.
        headless (bool): Whether to run in headless mode.

    Returns:
        Tuple[str, str]: The authorization URL and initial state.
    """

    kwargs = {'sign_up': sign_up, 'locale': 'en', 'ensure_auth': ensure_auth, 'headless': headless}
    if organization:
        kwargs['organization'] = organization.id

    return client.create_authorization_url(CLI_AUTH,
                                           code_verifier=code_verifier,
                                           **kwargs)

def get_logout_url(id_token: str) -> str:
    """
    Generate the logout URL.

    Args:
        id_token (str): The ID token.

    Returns:
        str: The logout URL.
    """
    return f'{CLI_AUTH_LOGOUT}?id_token={id_token}'

def get_redirect_url() -> str:
    """
    Get the redirect URL for the authentication callback.

    Returns:
        str: The redirect URL.
    """
    return CLI_CALLBACK

def get_organization() -> Optional[Organization]:
    """
    Retrieve the organization configuration.

    Returns:
        Optional[Organization]: The organization object, or None if not configured.
    """
    config = configparser.ConfigParser()
    config.read(CONFIG)

    org_conf: Union[Dict[str, str], configparser.SectionProxy] = config[
        'organization'] if 'organization' in config.sections() else {}
    org_id: Optional[str] = org_conf['id'].replace("\"", "") if org_conf.get('id', None) else None
    org_name: Optional[str] = org_conf['name'].replace("\"", "") if org_conf.get('name', None) else None

    if not org_id:
        return None

    org = Organization(
        id=org_id,
        name=org_name
    )

    return org

def get_auth_info(ctx) -> Optional[Dict]:
    """
    Retrieve the authentication information.

    Args:
        ctx: The context object containing authentication data.

    Returns:
        Optional[Dict]: The authentication information, or None if not authenticated.
    """
    from safety.auth.utils import is_email_verified

    info = None
    if ctx.obj.auth.client.token:
        try:
            info = get_token_data(get_token(name='id_token'), keys=ctx.obj.auth.keys)

            verified = is_email_verified(info)
            if not verified:
                user_info = ctx.obj.auth.client.fetch_user_info()
                verified = is_email_verified(user_info)

                if verified:
                    # refresh only if needed
                    raise ExpiredTokenError

        except ExpiredTokenError as e:
            # id_token expired. So fire a manually a refresh
            try:
                ctx.obj.auth.client.refresh_token(ctx.obj.auth.client.metadata.get('token_endpoint'),
                                         refresh_token=ctx.obj.auth.client.token.get('refresh_token'))
                info = get_token_data(get_token(name='id_token'), keys=ctx.obj.auth.keys)
            except Exception as _e:
                clean_session(ctx.obj.auth.client)
        except Exception as _g:
            clean_session(ctx.obj.auth.client)

    return info

def get_token_data(token: str, keys: Any, silent_if_expired: bool = False) -> Optional[Dict]:
    """
    Decode and validate the token data.

    Args:
        token (str): The token to decode.
        keys (Any): The keys to use for decoding.
        silent_if_expired (bool): Whether to silently ignore expired tokens.

    Returns:
        Optional[Dict]: The decoded token data, or None if invalid.
    """
    claims = jwt.decode(token, keys, claims_cls=CodeIDToken)
    try:
        claims.validate()
    except ExpiredTokenError as e:
        if not silent_if_expired:
            raise e

    return claims

def get_token(name: str = 'access_token') -> Optional[str]:
    """"
    Retrieve a token from the local authentication configuration.

    This returns tokens saved in the local auth configuration.
    There are two types of tokens: access_token and id_token

    Args:
        name (str): The name of the token to retrieve.

    Returns:
        Optional[str]: The token value, or None if not found.
    """
    config = configparser.ConfigParser()
    config.read(AUTH_CONFIG_USER)

    if 'auth' in config.sections() and name in config['auth']:
        value = config['auth'][name]
        if value:
            return value

    return None

def get_host_config(key_name: str) -> Optional[Any]:
    """
    Retrieve a configuration value from the host configuration.

    Args:
        key_name (str): The name of the configuration key.

    Returns:
        Optional[Any]: The configuration value, or None if not found.
    """
    config = configparser.ConfigParser()
    config.read(CONFIG)

    if not config.has_section("host"):
        return None

    host_section = dict(config.items("host"))

    if key_name in host_section:
        if key_name == "stage":
            # Support old alias in the config.ini
            if host_section[key_name] == "dev":
                host_section[key_name] = "development"
            if host_section[key_name] not in {env.value for env in Stage}:
                return None
            return Stage(host_section[key_name])

    return None

def str_to_bool(s: str) -> bool:
    """
    Convert a string to a boolean value.

    Args:
        s (str): The string to convert.

    Returns:
        bool: The converted boolean value.

    Raises:
        ValueError: If the string cannot be converted.
    """
    if s.lower() == 'true' or s == '1':
        return True
    elif s.lower() == 'false' or s == '0':
        return False
    else:
        raise ValueError(f"Cannot convert '{s}' to a boolean value.")

def get_proxy_config() -> Tuple[Optional[Dict[str, str]], Optional[int], bool]:
    """
    Retrieve the proxy configuration.

    Returns:
        Tuple[Optional[Dict[str, str]], Optional[int], bool]: The proxy configuration, timeout, and whether it is required.
    """
    config = configparser.ConfigParser()
    config.read(CONFIG)

    proxy_dictionary =  None
    required = False
    timeout = None
    proxy = None

    if config.has_section("proxy"):
        proxy = dict(config.items("proxy"))

    if proxy:
        try:
            proxy_dictionary = get_proxy_dict(proxy['protocol'], proxy['host'],
                                                proxy['port'])
            required = str_to_bool(proxy["required"])
            timeout = proxy["timeout"]
        except Exception as e:
            pass

    return proxy_dictionary, timeout, required

def clean_session(client) -> bool:
    """
    Clean the authentication session.

    Args:
        client: The authentication client.

    Returns:
        bool: Always returns True.
    """
    config = configparser.ConfigParser()
    config['auth'] = {'access_token': '', 'id_token': '', 'refresh_token':''}

    with open(AUTH_CONFIG_USER, 'w') as configfile:
        config.write(configfile)

    client.token = None

    return True

def save_auth_config(access_token: Optional[str] = None, id_token: Optional[str] = None, refresh_token: Optional[str] = None) -> None:
    """
    Save the authentication configuration.

    Args:
        access_token (Optional[str]): The access token.
        id_token (Optional[str]): The ID token.
        refresh_token (Optional[str]): The refresh token.
    """
    config = configparser.ConfigParser()
    config.read(AUTH_CONFIG_USER)
    config['auth'] = {'access_token': access_token, 'id_token': id_token,
                      'refresh_token': refresh_token}

    with open(AUTH_CONFIG_USER, 'w') as configfile:
        config.write(configfile)



import asyncio
import functools
import os
import platform
import re
import shutil
from typing import Dict, List, Optional, Set, Union


from safety_schemas.models.events.payloads import ToolStatus, AliasConfig, IndexConfig
from safety_schemas.models.events.types import ToolType


class ToolInspector:
    """
    Inspects the system for installed tools managers and their versions.
    """

    COMMON_LOCATIONS = {
        # Common paths across many tools
        "COMMON": ["/usr/local/bin", "/usr/bin", "~/.local/bin"],
        # Tool-specific paths
        ToolType.PIP: [
            # Virtual environments
            "venv/bin",
            "env/bin",
            ".venv/bin",
            ".env/bin",
            # Python installations
            "/opt/python*/bin",
            # Windows specific
            "C:/Python*/Scripts",
            "%APPDATA%/Python/Python*/Scripts",
            # macOS specific
            "/Library/Frameworks/Python.framework/Versions/*/bin",
        ],
        ToolType.POETRY: [
            "~/.poetry/bin",
            # Windows
            "%APPDATA%/Python/poetry/bin",
            "%USERPROFILE%/.poetry/bin",
        ],
        ToolType.CONDA: [
            "~/miniconda3/bin",
            "~/anaconda3/bin",
            "/opt/conda/bin",
            "/opt/miniconda3/bin",
            "/opt/anaconda3/bin",
            # Windows
            "C:/ProgramData/Miniconda3",
            "C:/ProgramData/Anaconda3",
            "%USERPROFILE%/Miniconda3",
            "%USERPROFILE%/Anaconda3",
        ],
        ToolType.UV: [
            "~/.cargo/bin",
            # Windows
            "%USERPROFILE%/.cargo/bin",
        ],
        ToolType.NPM: [
            "~/.nvm/versions/node/*/bin",
            # Windows
            "%APPDATA%/npm",
            "C:/Program Files/nodejs",
        ],
    }

    # Command arguments to check version
    VERSION_ARGS = {
        ToolType.PIP: "--version",
        ToolType.UV: "--version",
        ToolType.NPM: "--version",
        ToolType.POETRY: "--version",
        ToolType.CONDA: "--version",
    }

    # Version parsing regex
    VERSION_REGEX = {
        ToolType.PIP: r"pip (\d+\.\d+(?:\.\d+)?)",
        ToolType.UV: r"uv (\d+\.\d+(?:\.\d+)?)",
        ToolType.NPM: r"(\d+\.\d+\.\d+)",
        ToolType.POETRY: r"Poetry version (\d+\.\d+\.\d+)",
        ToolType.CONDA: r"conda (\d+\.\d+(?:\.\d+)?)",
    }

    def __init__(self, timeout: float = 1.0):
        """
        Initialize the detector.

        Args:
            timeout: Command execution timeout in seconds
        """
        self.timeout = timeout
        self._found_paths: Dict[ToolType, Set[str]] = {t: set() for t in ToolType}

    # TODO: limit concurrency
    async def inspect_all_tools(self) -> List[ToolStatus]:
        """
        Inspect all tools installed in the system.

        Returns:
            List of ToolStatus objects for each found tool
        """
        tasks = []
        for tool_type in ToolType:
            tasks.append(self._find_tool_instances(tool_type))

        results: List[Union[List[ToolStatus], BaseException]] = await asyncio.gather(
            *tasks, return_exceptions=True
        )

        tools_inspected: List[ToolStatus] = []

        for tool_status in results:
            if isinstance(tool_status, list):
                tools_inspected.extend(tool_status)

        return tools_inspected

    async def _find_tool_instances(self, tool_type: ToolType) -> List[ToolStatus]:
        """
        Find all instances of a specific tool type.
        """
        # Find all executable paths
        paths = await self._find_executable_paths(tool_type)

        tasks = [self._check_tool(tool_type, path) for path in paths]
        results: List[
            Optional[Union[ToolStatus, BaseException]]
        ] = await asyncio.gather(*tasks, return_exceptions=True)

        tools_inspected: List[ToolStatus] = []

        for tool_status in results:
            if isinstance(tool_status, ToolStatus):
                tools_inspected.append(tool_status)

        return tools_inspected

    def _search_executable_paths(self, tool_type: ToolType) -> Set[str]:
        # Get the executable name
        exe_name = tool_type.value
        if platform.system() == "Windows":
            exe_name = f"{exe_name}.exe"

        paths = set()

        path_result = shutil.which(exe_name)
        if path_result:
            paths.add(os.path.abspath(path_result))

        for location_pattern in (
            self.COMMON_LOCATIONS["COMMON"] + self.COMMON_LOCATIONS[tool_type]
        ):
            if location_pattern.startswith("~"):
                location_pattern = os.path.expanduser(location_pattern)

            if "%" in location_pattern:
                location_pattern = os.path.expandvars(location_pattern)

            # Handle wildcards
            if "*" in location_pattern:
                # This is a simplified wildcard expansion - a more robust implementation
                # would use glob or similar, but this is faster for common cases
                base_dir = location_pattern.split("*")[0]
                if os.path.exists(base_dir):
                    for root, dirs, files in os.walk(base_dir, followlinks=False):
                        if exe_name in files:
                            exe_path = os.path.join(root, exe_name)
                            if os.access(exe_path, os.X_OK):
                                paths.add(os.path.abspath(exe_path))
            else:
                # Direct path check
                exe_path = os.path.join(location_pattern, exe_name)
                if os.path.exists(exe_path) and os.access(exe_path, os.X_OK):
                    paths.add(os.path.abspath(exe_path))

        return paths

    async def _find_executable_paths(self, tool_type: ToolType) -> Set[str]:
        """
        Find all executable paths for a tool type.
        """
        if self._found_paths[tool_type]:
            return self._found_paths[tool_type]
        paths = await asyncio.get_event_loop().run_in_executor(
            None, functools.partial(self._search_executable_paths, tool_type)
        )
        self._found_paths[tool_type] = paths
        return paths

    def _kill_process(self, proc):
        """
        Helper method to kill a process safely.
        """
        if proc is None:
            return

        try:
            proc.kill()
        except Exception:
            # Ignore any errors during kill
            pass

    async def _check_tool(self, tool_type: ToolType, path: str) -> Optional[ToolStatus]:
        """
        Check if a tool at a specific path is reachable and get its version.
        """
        try:
            version_arg = self.VERSION_ARGS[tool_type]
            proc = await asyncio.create_subprocess_exec(
                path,
                version_arg,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

            try:
                stdout, stderr = await asyncio.wait_for(
                    proc.communicate(), timeout=self.timeout
                )
                output = stdout.decode() + stderr.decode()

                # Extract version
                version_match = re.search(self.VERSION_REGEX[tool_type], output)
                version = version_match.group(1) if version_match else "unknown"

                AliasConfig(is_configured=True)
                IndexConfig(is_configured=True)

                return ToolStatus(
                    type=tool_type,
                    command_path=path,
                    version=version,
                    reachable=proc.returncode == 0,
                    alias_config=None,
                    index_config=None,
                )
            except (asyncio.TimeoutError, TimeoutError):
                self._kill_process(proc)

                # Command timed out
                return ToolStatus(
                    type=tool_type,
                    command_path=path,
                    version="unknown",
                    reachable=False,
                )
        except Exception:
            # Any other error means the tool is not reachable
            return ToolStatus(
                type=tool_type, command_path=path, version="unknown", reachable=False
            )



import unittest
from unittest.mock import MagicMock
from safety.scan.ecosystems.python.main import should_fail, VulnerabilitySeverityLabels

class TestMain(unittest.TestCase):
    def setUp(self):
        self.config = MagicMock()
        self.vulnerability = MagicMock()

    def test_fail_on_disabled(self):
        self.config.depedendency_vulnerability.fail_on.enabled = False
        result = should_fail(self.config, self.vulnerability)
        self.assertFalse(result)

    def test_severity_none(self):
        self.config.depedendency_vulnerability.fail_on.enabled = True
        self.vulnerability.severity = None
        result = should_fail(self.config, self.vulnerability)
        self.assertFalse(result)

    def test_severity_none_with_fail_on_unknow_none(self):
        self.config.depedendency_vulnerability.fail_on.enabled = True        
        self.vulnerability.severity = None

        self.config.depedendency_vulnerability.fail_on.cvss_severity = [VulnerabilitySeverityLabels.UNKNOWN]
        self.assertTrue(should_fail(self.config, self.vulnerability))

        self.config.depedendency_vulnerability.fail_on.cvss_severity = [VulnerabilitySeverityLabels.NONE]
        self.assertTrue(should_fail(self.config, self.vulnerability))

        self.config.depedendency_vulnerability.fail_on.cvss_severity = [VulnerabilitySeverityLabels.UNKNOWN, 
                                                                        VulnerabilitySeverityLabels.NONE]
        self.assertTrue(should_fail(self.config, self.vulnerability))

        self.config.depedendency_vulnerability.fail_on.cvss_severity = [VulnerabilitySeverityLabels.LOW, 
                                                                        VulnerabilitySeverityLabels.MEDIUM]
        self.assertFalse(should_fail(self.config, self.vulnerability))

        self.vulnerability.severity = MagicMock()
        self.vulnerability.severity.cvssv3 = {"base_severity": "NONE"}
        
        self.config.depedendency_vulnerability.fail_on.cvss_severity = [VulnerabilitySeverityLabels.NONE]
        self.assertTrue(should_fail(self.config, self.vulnerability))

        self.config.depedendency_vulnerability.fail_on.cvss_severity = [VulnerabilitySeverityLabels.UNKNOWN]
        self.assertFalse(should_fail(self.config, self.vulnerability))

        self.vulnerability.severity.cvssv3 = {"base_severity": "UNKNOWN"}
        self.assertTrue(should_fail(self.config, self.vulnerability))

    def test_known_severity_failure(self):
        self.config.depedendency_vulnerability.fail_on.enabled = True
        self.config.depedendency_vulnerability.fail_on.cvss_severity = [VulnerabilitySeverityLabels.HIGH]
        self.vulnerability.severity.cvssv3 = {"base_severity": "HIGH"}
        result = should_fail(self.config, self.vulnerability)
        self.assertTrue(result)

    def test_unexpected_severity_with_warning(self):
        self.config.depedendency_vulnerability.fail_on.enabled = True
        self.config.depedendency_vulnerability.fail_on.cvss_severity = [VulnerabilitySeverityLabels.HIGH]
        self.vulnerability.severity.cvssv3 = {"base_severity": "UNKNOWN_SEVERITY"}
        with self.assertLogs(level='WARNING') as log:
            result = should_fail(self.config, self.vulnerability)
            self.assertIn("Unexpected base severity value", log.output[0])
        self.assertFalse(result)




import logging
import os
from pathlib import Path
import re
from typing import Dict, List, Optional, Set, Tuple, Union
from safety_schemas.models import Ecosystem, FileType

from safety.errors import SafetyException

from .handlers import FileHandler, ECOSYSTEM_HANDLER_MAPPING

LOG = logging.getLogger(__name__)

def should_exclude(excludes: Set[Path], to_analyze: Path) -> bool:
    """
    Determines whether a given path should be excluded based on the provided exclusion set.

    Args:
        excludes (Set[Path]): Set of paths to exclude.
        to_analyze (Path): The path to analyze.

    Returns:
        bool: True if the path should be excluded, False otherwise.
    """

    if not to_analyze.is_absolute():
        to_analyze = to_analyze.resolve()

    for exclude in excludes:
        if not exclude.is_absolute():
            exclude = exclude.resolve()

        try:
            if to_analyze == exclude or \
                to_analyze.relative_to(exclude):
                return True
        except ValueError:
            pass

    return False


class FileFinder():
    """"
    Defines a common interface to agree in what type of components Safety is trying to
    find depending on the language type.
    """

    def __init__(
        self,
        max_level: int,
        ecosystems: List[Ecosystem],
        target: Path,
        console,
        live_status=None,
        exclude: Optional[List[str]] = None,
        include_files: Optional[Dict[FileType, List[Path]]] = None,
        handlers: Optional[Set[FileHandler]] = None
    ) -> None:
        """
        Initializes the FileFinder with the specified parameters.

        Args:
            max_level (int): Maximum directory depth to search.
            ecosystems (List[Ecosystem]): List of ecosystems to consider.
            target (Path): Target directory to search.
            console: Console object for output.
            live_status: Live status object for updates.
            exclude (Optional[List[str]]): List of patterns to exclude from the search.
            include_files (Optional[Dict[FileType, List[Path]]]): Dictionary of files to include in the search.
            handlers (Optional[Set[FileHandler]]): Set of file handlers.
        """
        self.max_level = max_level
        self.target = target
        self.include_files = include_files

        # If no handlers are provided, initialize them from the ecosystem mapping
        if not handlers:
            handlers = set(ECOSYSTEM_HANDLER_MAPPING[ecosystem]()
                                              for ecosystem in ecosystems)

        self.handlers = handlers
        self.file_count = 0
        self.exclude_dirs: Set[Path] = set()
        self.exclude_files: Set[Path] = set()
        exclude = [] if not exclude else exclude

        # Populate the exclude_dirs and exclude_files sets based on the provided patterns
        for pattern in exclude:
            for path in Path(target).glob(pattern):
                if path.is_dir():
                    self.exclude_dirs.add(path)
                else:
                    self.exclude_files.add(path)

        self.console = console
        self.live_status = live_status

    def process_directory(self, dir_path: str, max_deep: Optional[int] = None) -> Tuple[str, Dict[str, Set[Path]]]:
        """
        Processes the specified directory to find files matching the handlers' criteria.

        Args:
            dir_path (str): The directory path to process.
            max_deep (Optional[int]): Maximum depth to search within the directory.

        Returns:
            Tuple[str, Dict[str, Set[Path]]]: The directory path and a dictionary of file types and their corresponding paths.
        """
        files: Dict[str, Set[Path]] = {}
        level : int = 0
        initial_depth = len(Path(dir_path).parts) - 1
        skip_dirs = set()
        skip_files = set()

        for root, dirs, filenames in os.walk(dir_path):
            root_path = Path(root)
            current_depth = len(root_path.parts) - initial_depth

            # Filter directories based on exclusion criteria
            dirs[:] = [d for d in dirs if not should_exclude(excludes=self.exclude_dirs,
                                                             to_analyze=(root_path / Path(d)))]
            if dirs:
                LOG.info(f"Directories to inspect -> {', '.join(dirs)}")

            LOG.info(f"Current -> {root}")
            if self.live_status:
                self.live_status.update(f":mag: Scanning {root}")

            # Stop descending into directories if the maximum depth is reached
            if max_deep is not None and current_depth > max_deep:
                # Don't go deeper
                del dirs[:]

            # Filter filenames based on exclusion criteria
            filenames[:] = [f for f in filenames if not should_exclude(
                excludes=self.exclude_files,
                to_analyze=Path(f))]

            self.file_count += len(filenames)

            for file_name in filenames:
                for handler in self.handlers:
                    file_type = handler.can_handle(root, file_name, self.include_files)
                    if file_type:
                        inspectable_file: Path = Path(root, file_name)
                        if file_type.value not in files or not files[file_type.value]:
                            files[file_type.value] = set()
                        files[file_type.value].add(inspectable_file)
                        break
            level += 1

        return dir_path, files

    def search(self) -> Tuple[str, Dict[str, Set[Path]]]:
        """
        Initiates the search for files within the target directory.

        Returns:
            Tuple[str, Dict[str, Set[Path]]]: The target directory and a dictionary of file types and their corresponding paths.
        """
        return self.process_directory(self.target, self.max_level)



import json
import unittest
from unittest.mock import patch

from safety.formatters.json import JsonReport
from packaging.version import parse

from safety.models import CVE, Vulnerability, Package, Severity, SafetyRequirement


class TestJSONFormatter(unittest.TestCase):

    def setUp(self) -> None:
        self.formatter = JsonReport()
        self.maxDiff = None

    @patch('safety.formatters.json.get_report_brief_info')
    def test_render_vulnerabilities_with_remediations(self, get_report_brief_info):
        get_report_brief_info.return_value = {'scan_target': 'environment',
                                              'scanned': ['/usr/local/lib/python3.9/site-packages'],
                                              'target_languages': ['python'],
                                              'api_key': True,
                                              'packages_found': 2,
                                              'timestamp': '2022-03-03 16:31:30',
                                              'safety_version': '2.0.0.dev6'}

        affected_package = Package(name='django', version='4.0.1',
                                   requirements=[SafetyRequirement('django==4.0.1',
                                                                   found='/usr/local/lib/python3.9/site-packages')],
                                   found='/usr/local/lib/python3.9/site-packages',
                                   insecure_versions=['4.0.1'], secure_versions=['4.0.4', '3.2.13', '2.2.28'],
                                   latest_version_without_known_vulnerabilities='',
                                   latest_version='4.0.4',
                                   more_info_url='https://pyup.io/packages/pypi/django/?from=4.0.1&to=4.0.4')
        remediations = {
            'django': {'==4.0.1': {'vulnerabilities_found': 1, 'version': '4.0.1',
                                   'requirement': SafetyRequirement('django==4.0.1'),
                                   'recommended_version': parse('4.0.4'),
                                   'other_recommended_versions': ['2.2.28', '3.2.13'],                                   
                                   'closest_secure_version': {'upper': parse('4.0.4'),
                                                              'lower': None},
                                   'more_info_url': 'https://pyup.io/packages/pypi/django/?from=4.0.1&to=4.0.4'}}}
        cve = CVE(name='CVE-2022-22818',
                  cvssv2={'base_score': 4.3, 'impact_score': 2.9,
                          'vector_string': 'AV:N/AC:M/Au:N/C:N/I:P/A:N'},
                  cvssv3={'base_score': 6.1, 'impact_score': 2.7,
                          'base_severity': 'MEDIUM',
                          'vector_string': 'CVSS:3.1/AV:N/AC:L/PR:N/UI:R/S:C/C:L/I:L/A:N'})
        vulnerabilities = [Vulnerability(package_name='django', pkg=affected_package,
                                         ignored=False, ignored_reason=None, ignored_expires=None, vulnerable_spec='>=4.0a1,<4.0.2',
                                         all_vulnerable_specs=['>=4.0a1,<4.0.2'],
                                         analyzed_version='4.0.1',
                                         analyzed_requirement=SafetyRequirement('django==4.0.1'),
                                         advisory='The {% debug %} template tag in Django',
                                         vulnerability_id='44742', is_transitive=False, published_date='2022-Feb-03',
                                         fixed_versions=['2.2.27', '3.2.12', '4.0.2'],
                                         closest_versions_without_known_vulnerabilities=[],
                                         resources=['https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-2022-22818'],
                                         CVE=cve,
                                         severity=Severity(source=cve.name, cvssv2=cve.cvssv2, cvssv3=cve.cvssv3),
                                         affected_versions=['4.0.1'],
                                         more_info_url='https://pyup.io/vulnerabilities/CVE-2022-22818/44742/')]
        packages = [Package(name='secure-package', version='0.1.0',
                            requirements=[SafetyRequirement('secure-package==0.1.0')],
                            found='/usr/local/lib/python3.9/site-packages',
                            insecure_versions=[], secure_versions=['0.1.0'],
                            latest_version_without_known_vulnerabilities='',
                            latest_version='0.1.0',
                            more_info_url='https://pyup.io/packages/pypi/secure-package/'),
                    affected_package]

        result = self.formatter.render_vulnerabilities(announcements=[], vulnerabilities=vulnerabilities,
                                                       remediations=remediations, full=True, packages=packages,
                                                       fixes=[])

        expected_result = json.dumps(
            {
                "report_meta": {
                    "scan_target": "environment",
                    "scanned": [
                        "/usr/local/lib/python3.9/site-packages"
                    ],
                    "target_languages": [
                        "python"
                    ],
                    "api_key": True,
                    "packages_found": 2,
                    "timestamp": "2022-03-03 16:31:30",
                    "safety_version": "2.0.0.dev6",
                    "remediations_attempted": 0,
                    "remediations_completed": 0,
                    "remediation_mode": "NON_INTERACTIVE"
                },
                "scanned_packages": {
                    "secure-package": {
                        "name": "secure-package",
                        "version": "0.1.0",
                        "requirements": [
                            {
                                "raw": "secure-package==0.1.0",
                                "extras": [],
                                "marker": None,
                                "name": "secure-package",
                                "specifier": "==0.1.0",
                                "url": None,
                                "found": None
                            },
                        ]
                    },
                    "django": {
                        "name": "django",
                        "version": "4.0.1",
                        "requirements": [
                            {
                                "raw": "django==4.0.1",
                                "extras": [],
                                "marker": None,
                                "name": "django",
                                "specifier": "==4.0.1",
                                "url": None,
                                "found": "/usr/local/lib/python3.9/site-packages"
                            }
                        ]
                    },
                },
                "affected_packages": {
                    "django": {
                        "name": "django",
                        "version": "4.0.1",
                        "requirements":[
                            {
                                "raw": "django==4.0.1",
                                "extras": [],
                                "marker": None,
                                "name": "django",
                                "specifier": "==4.0.1",
                                "url": None,
                                "found": "/usr/local/lib/python3.9/site-packages"
                            }
                        ],
                        "found": None,
                        "insecure_versions": [
                            "4.0.1"
                        ],
                        "secure_versions": [
                            "4.0.4",
                            "3.2.13",
                            "2.2.28"
                        ],
                        "latest_version_without_known_vulnerabilities": "",
                        "latest_version": "4.0.4",
                        "more_info_url": "https://pyup.io/packages/pypi/django/?from=4.0.1&to=4.0.4"
                    }
                },
                "announcements": [],
                "vulnerabilities": [
                    {
                        "vulnerability_id": "44742",
                        "package_name": "django",
                        "ignored": False,
                        "ignored_reason": None,
                        "ignored_expires": None,
                        "vulnerable_spec": ">=4.0a1,<4.0.2",
                        "all_vulnerable_specs": [
                            ">=4.0a1,<4.0.2"
                        ],
                        "analyzed_version": "4.0.1",
                        "analyzed_requirement": {
                            "raw": "django==4.0.1",
                            "extras": [],
                            "marker": None,
                            "name": "django",
                            "specifier": "==4.0.1",
                            "url": None,
                            "found": None
                        },
                        "advisory": "The {% debug %} template tag in Django",
                        "is_transitive": False,
                        "published_date": "2022-Feb-03",
                        "fixed_versions": [
                            "2.2.27",
                            "3.2.12",
                            "4.0.2"
                        ],
                        "closest_versions_without_known_vulnerabilities": [],
                        "resources": [
                            "https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-2022-22818"
                        ],
                        "CVE": "CVE-2022-22818",
                        "severity": {
                            "source": "CVE-2022-22818",
                            "cvssv2": {
                                "base_score": 4.3,
                                "impact_score": 2.9,
                                "vector_string": "AV:N/AC:M/Au:N/C:N/I:P/A:N"
                            },
                            "cvssv3": {
                                "base_score": 6.1,
                                "impact_score": 2.7,
                                "base_severity": "MEDIUM",
                                "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:R/S:C/C:L/I:L/A:N"
                            }
                        },
                        "affected_versions": [
                            "4.0.1",
                        ],
                        "more_info_url": "https://pyup.io/vulnerabilities/CVE-2022-22818/44742/"
                    },
                ],
                "ignored_vulnerabilities": [],
                "remediations": {
                    "django": {
                        "requirements": {
                            "==4.0.1": {
                                "vulnerabilities_found": 1,
                                "version": "4.0.1",
                                "requirement": {
                                    "raw": "django==4.0.1",
                                    "extras": [],
                                    "marker": None,
                                    "name": "django",
                                    "specifier": "==4.0.1",
                                    "url": None,
                                    "found": None
                                },
                                "recommended_version": "4.0.4",
                                "other_recommended_versions": [
                                    "2.2.28",
                                    "3.2.13"
                                ],
                                "closest_secure_version": {
                                    "upper": "4.0.4",
                                    "lower": None
                                },
                                "more_info_url": "https://pyup.io/packages/pypi/django/?from=4.0.1&to=4.0.4"
                            }
                            },
                        "current_version": None,
                        "vulnerabilities_found": None,
                        "recommended_version": None,
                        "other_recommended_versions": [],
                        "more_info_url": None
                    }
                },
                "remediations_results": {
                    "vulnerabilities_fixed": [],
                    "remediations_applied": {},
                    "remediations_skipped": {}
                }
            },
            indent=4
        )
        self.assertEqual(result, expected_result)



from pathlib import Path

from safety.constants import USER_CONFIG_DIR, get_config_setting

AUTH_CONFIG_FILE_NAME = "auth.ini"
AUTH_CONFIG_USER = USER_CONFIG_DIR / Path(AUTH_CONFIG_FILE_NAME)


HOST: str = 'localhost'

CLIENT_ID = get_config_setting("CLIENT_ID")
AUTH_SERVER_URL = get_config_setting("AUTH_SERVER_URL")
SAFETY_PLATFORM_URL = get_config_setting("SAFETY_PLATFORM_URL")

OPENID_CONFIG_URL = f"{AUTH_SERVER_URL}/.well-known/openid-configuration"

CLAIM_EMAIL_VERIFIED_API = "https://api.safetycli.com/email_verified"
CLAIM_EMAIL_VERIFIED_AUTH_SERVER = "email_verified"

CLI_AUTH = f'{SAFETY_PLATFORM_URL}/cli/auth'
CLI_AUTH_SUCCESS = f'{SAFETY_PLATFORM_URL}/cli/auth/success'
CLI_AUTH_LOGOUT = f'{SAFETY_PLATFORM_URL}/cli/logout'
CLI_CALLBACK = f'{SAFETY_PLATFORM_URL}/cli/callback'
CLI_LOGOUT_SUCCESS = f'{SAFETY_PLATFORM_URL}/cli/logout/success'

MSG_NON_AUTHENTICATED = "Safety is not authenticated. Please run 'safety auth login' to log in."
MSG_FAIL_LOGIN_AUTHED = """[green]You are authenticated as[/green] {email}.

To log into a different account, first logout via: safety auth logout, and then login again."""
MSG_FAIL_REGISTER_AUTHED = 'You are currently logged in to {email}, please logout using `safety auth logout` before registering a new account.'

MSG_LOGOUT_DONE = "[green]Logout done.[/green]"
MSG_LOGOUT_FAILED = "[red]Logout failed. Try again.[/red]"


import base64
import json
import shutil
import subprocess
import urllib
from pathlib import Path
from typing import Optional
from urllib.parse import urlsplit, urlunsplit, urlencode

import typer
from rich.console import Console

from safety.tool.constants import PUBLIC_REPOSITORY_URL, ORGANIZATION_REPOSITORY_URL
from safety.tool.resolver import get_unwrapped_command

from safety.console import main_console


class Pip:
    @classmethod
    def is_installed(cls) -> bool:
        """
        Checks if the PIP program is installed

        Returns:
            True if PIP is installed on system, or false otherwise
        """
        return shutil.which("pip") is not None

    @classmethod
    def configure_requirements(
        cls,
        file: Path,
        org_slug: Optional[str],
        project_id: str,
        console: Console = main_console,
    ) -> None:
        """
        Configures Safety index url for specified requirements file.

        Args:
            file (Path): Path to requirements.txt file.
            org_slug (str): Organization slug.
            project_id (str): Project id.
            console (Console): Console instance.
        """

        with open(file, "r+") as f:
            content = f.read()

            repository_url = (
                ORGANIZATION_REPOSITORY_URL.format(org_slug)
                if org_slug
                else PUBLIC_REPOSITORY_URL
            )
            repository_url = (
                repository_url
                + "?"
                + urllib.parse.urlencode({"project-id": project_id})  # type: ignore
            )
            index_config = f"-i {repository_url}\n"
            if content.find(index_config) == -1:
                f.seek(0)
                f.write(index_config + content)

                console.print(f"Configured {file} file")
            else:
                console.print(f"{file} is already configured. Skipping.")

    @classmethod
    def configure_system(cls, org_slug: Optional[str], console: Console = main_console):
        """
        Configures PIP system to use to Safety index url.
        """
        try:
            repository_url = (
                ORGANIZATION_REPOSITORY_URL.format(org_slug)
                if org_slug
                else PUBLIC_REPOSITORY_URL
            )
            subprocess.run(
                [
                    get_unwrapped_command(name="pip"),
                    "config",
                    "set",
                    "global.index-url",
                    repository_url,
                ],
                capture_output=True,
            )
            console.print("Configured PIP global settings")
        except Exception:
            console.print("Failed to configure PIP global settings.")

    @classmethod
    def reset_system(cls, console: Console = main_console):
        # TODO: Move this logic and implement it in a more robust way
        try:
            subprocess.run(
                [
                    get_unwrapped_command(name="pip"),
                    "config",
                    "unset",
                    "global.index-url",
                ],
                capture_output=True,
            )
        except Exception:
            console.print("Failed to reset PIP global settings.")

    @classmethod
    def index_credentials(cls, ctx: typer.Context):
        auth_envelop = json.dumps(
            {
                "version": "1.0",
                "access_token": ctx.obj.auth.client.token["access_token"],
                "api_key": ctx.obj.auth.client.api_key,
                "project_id": ctx.obj.project.id if ctx.obj.project else None,
            }
        )
        return base64.urlsafe_b64encode(auth_envelop.encode("utf-8")).decode("utf-8")

    @classmethod
    def default_index_url(cls) -> str:
        return "https://pypi.org/simple/"

    @classmethod
    def build_index_url(cls, ctx: typer.Context, index_url: Optional[str]) -> str:
        if index_url is None:
            index_url = (
                PUBLIC_REPOSITORY_URL
                + "?"
                + urlencode({"project-id": ctx.obj.project.id})
            )

        url = urlsplit(index_url)

        encoded_auth = cls.index_credentials(ctx)
        netloc = f"user:{encoded_auth}@{url.netloc}"

        if type(url.netloc) is bytes:
            url = url._replace(netloc=netloc.encode("utf-8"))
        elif type(url.netloc) is str:
            url = url._replace(netloc=netloc)

        return urlunsplit(url)



from pathlib import Path
import unittest
from unittest.mock import Mock, call, patch

from safety.init.main import (
    PROJECT_CONFIG_ID,
    PROJECT_CONFIG_NAME,
    PROJECT_CONFIG_URL,
    PROJECT_CONFIG_SECTION,
    check_project,
    create_project,
    save_project_info,
    save_verified_project,
)

from safety_schemas.models import ProjectModel, Stage


class TestInitMain(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    @patch("safety.init.main.prompt_project_id")
    @patch("safety.init.main.print_wait_project_verification")
    def test_check_project_without_id(self, mock_wait_verification, mock_prompt_id):
        """
        If not project id is provided, the user should be prompted for one,
        then the project should be verified.
        """
        ctx = Mock()
        ctx.obj.auth.stage = Stage.production
        ctx.obj.telemetry = Mock(safety_source="cli")

        session = Mock()
        console = Mock()
        unverified_project = Mock(
            id=None, project_path=Path("/test/dir/project/.safety-project.ini")
        )

        # Mock prompt returning project id
        mock_prompt_id.return_value = "prompted-id"
        mock_wait_verification.return_value = {"status": "success"}

        _ = check_project(ctx, session, console, unverified_project, git_origin=None)

        # Assert prompt was called with parent dir name
        mock_prompt_id.assert_called_once_with(console, "project")

        # Assert core data is correct with prompted id
        expected_data = {
            "scan_stage": Stage.production,
            "safety_source": "cli",
            "project_slug": "prompted-id",
            "project_slug_source": "user",
        }

        # Assert verification called with prompted data
        mock_wait_verification.assert_called_once_with(
            console,
            "prompted-id",
            (session.check_project, expected_data),
            on_error_delay=1,
        )

    @patch("safety.init.main.save_project_info")
    def test_save_verified_project(self, mock_save_project_info):
        ctx = Mock()
        ctx.obj = Mock()

        project = {
            "id": "test-project",
            "name": "Test Project",
            "project_path": Path("/path/to/project"),
            "url_path": "/test/url",
        }

        organization = {"name": "Test Organization", "slug": "test-org"}

        save_verified_project(
            ctx,
            slug=project["id"],
            **{
                k: v
                for k, v in {**project, **{"organization": organization}}.items()
                if k != "id"
            },
        )

        # Assert project is correct type and values
        self.assertIsInstance(ctx.obj.project, ProjectModel)
        for attr, expected in project.items():
            self.assertEqual(getattr(ctx.obj.project, attr), expected)

        self.assertEqual(ctx.obj.org.get("name"), organization.get("name"))
        self.assertEqual(ctx.obj.org.get("slug"), organization.get("slug"))

        mock_save_project_info.assert_called_once_with(
            project=ctx.obj.project, project_path=project["project_path"]
        )

    @patch("configparser.ConfigParser")
    def test_save_project_info_success(self, mock_config_cls):
        # case_name, project, expected_config
        test_cases = [
            (
                "full_project",
                ProjectModel(
                    id="test-id", url_path="http://example.com", name="Test Project"
                ),
                {
                    PROJECT_CONFIG_ID: "test-id",
                    PROJECT_CONFIG_URL: "http://example.com",
                    PROJECT_CONFIG_NAME: "Test Project",
                },
            ),
            ("id_only", ProjectModel(id="test-id"), {PROJECT_CONFIG_ID: "test-id"}),
            (
                "with_url",
                ProjectModel(id="test-id", url_path="http://example.com"),
                {
                    PROJECT_CONFIG_ID: "test-id",
                    PROJECT_CONFIG_URL: "http://example.com",
                },
            ),
            (
                "with_name",
                ProjectModel(id="test-id", name="Test Project"),
                {PROJECT_CONFIG_ID: "test-id", PROJECT_CONFIG_NAME: "Test Project"},
            ),
        ]

        mock_config = mock_config_cls.return_value
        section_mock = mock_config.__getitem__.return_value

        for case_name, project, expected_config in test_cases:
            with self.subTest(case_name=case_name):
                mock_config.reset_mock()

                result = save_project_info(project, "test_config.ini")  # type: ignore

                # Assert the result
                self.assertTrue(result)
                mock_config.__getitem__.assert_called_with(PROJECT_CONFIG_SECTION)

                calls = [call(key, value) for key, value in expected_config.items()]
                section_mock.__setitem__.assert_has_calls(calls, any_order=False)

    @patch("configparser.ConfigParser")
    def test_save_project_info_file_error(self, mock_config_cls):
        project = ProjectModel(id="test-id")

        mock_config = mock_config_cls.return_value
        mock_config.write.side_effect = Exception("Write error")

        result = save_project_info(project, "test_config.ini")  # type: ignore

        self.assertFalse(result)

    @patch("safety.init.main.load_unverified_project_from_config")
    @patch("safety.init.main.GIT")
    @patch("safety.init.main.verify_project")
    def test_create_project_with_platform_enabled(
        self, mock_verify, mock_git, mock_load_project
    ):
        ctx = Mock()
        ctx.obj.platform_enabled = True
        console = Mock()
        target = Path("/some/path")

        mock_git_instance = mock_git.return_value
        mock_git_instance.build_git_data.return_value = Mock(origin="test-origin")
        mock_project = mock_load_project.return_value

        create_project(ctx, console, target)

        # Make sure project loads unverified, git data is built and
        # project verification is called
        mock_load_project.assert_called_once_with(project_root=target)
        mock_git_instance.build_git_data.assert_called_once()
        mock_verify.assert_called_once_with(
            console,
            ctx,
            ctx.obj.auth.client,
            mock_project,
            ctx.obj.auth.stage,
            "test-origin",
        )

    @patch("safety.init.main.load_unverified_project_from_config")
    @patch("safety.init.main.GIT")
    @patch("safety.init.main.verify_project")
    def test_create_project_platform_disabled(
        self, mock_verify, mock_git, mock_load_project
    ):
        ctx = Mock()
        ctx.obj.platform_enabled = False
        console = Mock()
        target = Path("/some/path")

        create_project(ctx, console, target)

        mock_verify.assert_not_called()
        console.print.assert_called_once_with(
            "Project creation is not supported for your account."
        )



import os.path
from pathlib import Path
from typing import Optional

from safety.console import main_console as console
from safety.tool.utils import (
    PipConfigurator,
    PipRequirementsConfigurator,
    PoetryPyprojectConfigurator,
    is_os_supported,
)

from .interceptors import create_interceptor


def has_local_tool_files(directory: Path) -> bool:
    configurators = [PipRequirementsConfigurator(), PoetryPyprojectConfigurator()]

    for file_name in os.listdir(directory):
        if os.path.isfile(file_name):
            file = Path(file_name)
            for configurator in configurators:
                if configurator.is_supported(file):
                    return True

    return False


def configure_system(org_slug: Optional[str]):
    configurators = [PipConfigurator()]

    for configurator in configurators:
        configurator.configure(org_slug)


def reset_system():
    configurators = [PipConfigurator()]

    for configurator in configurators:
        configurator.reset()


def configure_alias():
    if not is_os_supported():
        return

    interceptor = create_interceptor()
    interceptor.install_interceptors()

    console.print("Configured PIP alias")


def configure_local_directory(
    directory: Path, org_slug: Optional[str], project_id: Optional[str]
):
    configurators = [PipRequirementsConfigurator(), PoetryPyprojectConfigurator()]

    for file_name in os.listdir(directory):
        if os.path.isfile(file_name):
            file = Path(file_name)
            for configurator in configurators:
                if configurator.is_supported(file):
                    configurator.configure(file, org_slug, project_id)



from rich.console import Console

from safety_schemas.models import Vulnerability, RemediationModel
from safety.scan.render import get_render_console
console = Console()

Vulnerability.__render__ = get_render_console(Vulnerability)



from .cli_utils import auth_options, build_client_session, proxy_options, \
    inject_session
from .cli import auth

__all__ = [
    "build_client_session",
    "proxy_options",
    "auth_options",
    "inject_session",
    "auth"
]


from unittest.mock import patch, ANY
from click.testing import CliRunner
import unittest

from safety.cli import cli

class TestSafetyAuthCLI(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.runner = CliRunner(mix_stderr=False)

        cli.commands = cli.all_commands
        self.cli = cli

    @unittest.skip("We are bypassing email verification for now")
    @patch("safety.auth.cli.fail_if_authenticated")
    @patch("safety.auth.cli.get_authorization_data")
    @patch("safety.auth.cli.process_browser_callback")
    def test_auth_calls_login(
        self, process_browser_callback, get_authorization_data, fail_if_authenticated
    ):
        auth_data = "https://safetycli.com", "initialState"
        get_authorization_data.return_value = auth_data
        process_browser_callback.return_value = {
            "email": "user@safetycli.com",
            "name": "Safety User",
        }
        result = self.runner.invoke(self.cli, ["auth"])

        fail_if_authenticated.assert_called_once()
        get_authorization_data.assert_called_once()
        process_browser_callback.assert_called_once_with(
            auth_data[0], initial_state=auth_data[1], ctx=ANY, headless=False
        )

        expected = [
            "",
            "Redirecting your browser to log in; once authenticated, return here to start using Safety",
            "",
            "You're authenticated",
            " Account: Safety User, user@safetycli.com (email verification required)",
            "",
            "To complete your account open the “verify your email” email sent to",
            "user@safetycli.com",
            "",
            "Can’t find the verification email? Login at",
            "`https://platform.safetycli.com/login/` to resend the verification email",
            "",
        ]

        for res_line, exp_line in zip(result.stdout.splitlines(), expected):
            self.assertIn(exp_line, res_line)



import datetime
import itertools
import json
import logging
import time
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import typer
from rich.console import Console
from rich.padding import Padding
from rich.prompt import Prompt
from rich.text import Text
from safety_schemas.models import (
    Ecosystem,
    FileType,
    IgnoreCodes,
    PolicyFileModel,
    PolicySource,
    ProjectModel,
    PythonDependency,
    ReportModel,
    Vulnerability,
)

from safety import safety
from safety.auth.constants import SAFETY_PLATFORM_URL
from safety.errors import SafetyException
from safety.meta import get_version
from safety.output_utils import parse_html
from safety.scan.constants import DEFAULT_SPINNER
from safety.util import clean_project_id, get_basic_announcements

LOG = logging.getLogger(__name__)


def render_header(targets: List[Path], is_system_scan: bool) -> Text:
    """
    Render the header text for the scan.

    Args:
        targets (List[Path]): List of target paths for the scan.
        is_system_scan (bool): Indicates if the scan is a system scan.

    Returns:
        Text: Rendered header text.
    """
    version = get_version()
    scan_datetime = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")

    action = f"scanning {', '.join([str(t) for t in targets])}"
    if is_system_scan:
        action = "running [bold]system scan[/bold]"

    return Text.from_markup(
        f"[bold]Safety[/bold] {version} {action}\n{scan_datetime}")

def print_header(console, targets: List[Path], is_system_scan: bool = False) -> None:
    """
    Print the header for the scan.

    Args:
        console (Console): The console for output.
        targets (List[Path]): List of target paths for the scan.
        is_system_scan (bool): Indicates if the scan is a system scan.
    """
    console.print(render_header(targets, is_system_scan), markup=True)

def print_announcements(console: Console, ctx: typer.Context):
    """
    Print announcements from Safety.

    Args:
        console (Console): The console for output.
        ctx (typer.Context): The context of the Typer command.
    """
    colors = {"error": "red", "warning": "yellow", "info": "default"}

    announcements = safety.get_announcements(ctx.obj.auth.client,
                                             telemetry=ctx.obj.config.telemetry_enabled,
                                             with_telemetry=ctx.obj.telemetry)
    basic_announcements = get_basic_announcements(announcements, False)

    if any(basic_announcements):
        console.print()
        console.print("[bold]Safety Announcements:[/bold]")
        console.print()
        for announcement in announcements:
            color = colors.get(announcement.get('type', "info"), "default")
            console.print(f"[{color}]* {announcement.get('message')}[/{color}]")

def print_detected_ecosystems_section(console: Console, file_paths: Dict[str, Set[Path]], include_safety_prjs: bool = True) -> None:
    """
    Print detected ecosystems section.

    Args:
        console (Console): The console for output.
        file_paths (Dict[str, Set[Path]]): Dictionary of file paths by type.
        include_safety_prjs (bool): Whether to include safety projects.
    """
    detected: Dict[Ecosystem, Dict[FileType, int]] = {}

    for file_type_key, f_paths in file_paths.items():
        file_type = FileType(file_type_key)
        if file_type.ecosystem:
            if file_type.ecosystem not in detected:
                detected[file_type.ecosystem] = {}
            detected[file_type.ecosystem][file_type] = len(f_paths)

    for ecosystem, f_type_count in detected.items():

        if not include_safety_prjs and ecosystem is Ecosystem.SAFETY_PROJECT:
            continue

        brief = "Found "
        file_types = []

        for f_type, count in f_type_count.items():
            file_types.append(f"{count} {f_type.human_name(plural=count>1)}")

        if len(file_types) > 1:
            brief += ", ".join(file_types[:-1]) + " and " + file_types[-1]
        else:
            brief += file_types[0]

        msg = f"{ecosystem.name.replace('_', ' ').title()} detected. {brief}"

        console.print(msg)
        


def print_fixes_section(console: Console, requirements_txt_found: bool = False, is_detailed_output: bool = False) -> None:
    """
    Print the section on applying fixes.

    Args:
        console (Console): The console for output.
        requirements_txt_found (bool): Indicates if a requirements.txt file was found.
        is_detailed_output (bool): Indicates if detailed output is enabled.
    """
    console.print("-" * console.size.width)
    console.print("Apply Fixes")
    console.print("-" * console.size.width)

    console.print()

    if requirements_txt_found:
        console.print("[green]Run `safety scan --apply-fixes`[/green] to update these packages and fix these vulnerabilities. "
                    "Documentation, limitations, and configurations for applying automated fixes: [link]https://docs.safetycli.com/safety-docs/vulnerability-remediation/applying-fixes[/link]")
        console.print()
        console.print("Alternatively, use your package manager to update packages to their secure versions. Always check for breaking changes when updating packages.")
    else:
        msg = "Use your package manager to update packages to their secure versions. Always check for breaking changes when updating packages."
        console.print(msg)

    if not is_detailed_output:
        console.print("[tip]Tip[/tip]: For more detailed output on each vulnerability, add the `--detailed-output` flag to safety scan.")

    console.print()
    console.print("-" * console.size.width)

def print_summary(
    console: Console,
    total_issues_with_duplicates: int,
    total_ignored_issues: int,
    project: ProjectModel,
    dependencies_count: int = 0,
    fixes_count: int = 0,
    resolved_vulns_per_fix: int = 0,
    is_detailed_output: bool = False,
    ignored_vulns_data: Optional[Dict[str, Vulnerability]] = None
) -> None:
    """
    Prints a concise summary of scan results including vulnerabilities, fixes, and ignored vulnerabilities.

    This function summarizes the results of a security scan, displaying the number of dependencies scanned,
    vulnerabilities found, suggested fixes, and the impact of those fixes. It also optionally provides a 
    detailed breakdown of ignored vulnerabilities based on predefined policies.

    Args:
        console (Console): The console object used to print formatted output.
        total_issues_with_duplicates (int): The total number of security issues, including duplicates.
        total_ignored_issues (int): The number of issues that were ignored based on project policies.
        project (ProjectModel): The project model containing the scanned project details and policies.
        dependencies_count (int, optional): The total number of dependencies scanned for vulnerabilities. Defaults to 0.
        fixes_count (int, optional): The number of fixes suggested by the scan. Defaults to 0.
        resolved_vulns_per_fix (int, optional): The number of vulnerabilities that can be resolved by the suggested fixes. Defaults to 0.
        is_detailed_output (bool, optional): Flag to indicate whether detailed output, especially for ignored vulnerabilities, should be shown. Defaults to False.
        ignored_vulns_data (Optional[Dict[str, Vulnerability]], optional): A dictionary of vulnerabilities that were ignored, categorized by their reason for being ignored. Defaults to None.

    Returns:
        None: This function does not return any value. It prints the summary to the console.

    Usage:
        Call this function after a vulnerability scan to display the results in a clear, formatted manner.
        Example:
            print_summary(console, unique_issues, 10, 2, project_model, dependencies_count=5, fixes_count=2)

    """
    
    from ..util import pluralize

    # Set the policy message based on the project source
    if project.policy:
        policy_msg = "policy fetched from Safety Platform" if project.policy.source is PolicySource.cloud else f"local {project.id or 'scan policy file'} project scan policy"
    else:
        policy_msg = "default Safety CLI policies"

    console.print(f"Tested [number]{dependencies_count}[/number] {pluralize('dependency', dependencies_count)} for security issues using {policy_msg}")

    if total_issues_with_duplicates == 0:
        console.print("0 security issues found, 0 fixes suggested.")
    else:
        # Print security issues and ignored vulnerabilities
        console.print(f"[number]{total_issues_with_duplicates}[/number] {pluralize('vulnerability', total_issues_with_duplicates)} found, "
                      f"[number]{total_ignored_issues}[/number] ignored due to policy.")

    console.print(
                  f"[number]{fixes_count}[/number] {pluralize('fix', fixes_count)} suggested, resolving [number]{resolved_vulns_per_fix}[/number] vulnerabilities.")
    
    if is_detailed_output:
        if not ignored_vulns_data:
            ignored_vulns_data = iter([])


        manual_ignored = {}
        cvss_severity_ignored = {}
        cvss_severity_ignored_pkgs = set()
        unpinned_ignored = {}
        unpinned_ignored_pkgs = set()
        environment_ignored = {}
        environment_ignored_pkgs = set()

        for vuln_data in ignored_vulns_data:
            code = IgnoreCodes(vuln_data.ignored_code)
            if code is IgnoreCodes.manual:
                manual_ignored[vuln_data.vulnerability_id] = vuln_data
            elif code is IgnoreCodes.cvss_severity:
                cvss_severity_ignored[vuln_data.vulnerability_id] = vuln_data
                cvss_severity_ignored_pkgs.add(vuln_data.package_name)
            elif code is IgnoreCodes.unpinned_specification:
                unpinned_ignored[vuln_data.vulnerability_id] = vuln_data
                unpinned_ignored_pkgs.add(vuln_data.package_name)
            elif code is IgnoreCodes.environment_dependency:
                environment_ignored[vuln_data.vulnerability_id] = vuln_data
                environment_ignored_pkgs.add(vuln_data.package_name)

        if manual_ignored:
            count = len(manual_ignored)
            console.print(
                f"[number]{count}[/number] were manually ignored due to the project policy:")
            for vuln in manual_ignored.values():
                render_to_console(vuln, console,
                                    rich_kwargs={"emoji": True, "overflow": "crop"},
                                    detailed_output=is_detailed_output)
        if cvss_severity_ignored:
            count = len(cvss_severity_ignored)
            console.print(
                f"[number]{count}[/number] {pluralize('vulnerability', count)} {pluralize('was', count)} ignored because " \
                        "of their severity or exploitability impacted the following" \
                        f" {pluralize('package', len(cvss_severity_ignored_pkgs))}: {', '.join(cvss_severity_ignored_pkgs)}"
            )
            
        if environment_ignored:
            count = len(environment_ignored)
            console.print(
                f"[number]{count}[/number] {pluralize('vulnerability', count)} {pluralize('was', count)} ignored because " \
                        "they are inside an environment dependency."
            )
            
        if unpinned_ignored:
            count = len(unpinned_ignored)
            console.print(
                f"[number]{count}[/number] {pluralize('vulnerability', count)} {pluralize('was', count)} ignored because " \
                    f"{pluralize('this', len(unpinned_ignored_pkgs))} {pluralize('package', len(unpinned_ignored_pkgs))} {pluralize('has', len(unpinned_ignored_pkgs))} unpinned specs: " \
                        f"{', '.join(unpinned_ignored_pkgs)}"
            )

       

def print_wait_project_verification(console: Console, project_id: str, closure: Tuple[Any, Dict[str, Any]], on_error_delay: int = 1) -> Any:
    """
    Print a waiting message while verifying a project.

    Args:
        console (Console): The console for output.
        project_id (str): The project ID.
        closure (Tuple[Any, Dict[str, Any]]): The function and its arguments to call.
        on_error_delay (int): Delay in seconds on error.

    Returns:
        Any: The status of the project verification.
    """
    status = None
    wait_msg = f"Verifying project {project_id} with Safety Platform."

    with console.status(wait_msg, spinner=DEFAULT_SPINNER):
        try:
            f, kwargs = closure
            status = f(**kwargs)
        except Exception as e:
            LOG.exception(f'Unable to verify the project, reason: {e}')
            reason = "We are currently unable to verify the project, " \
                "and it is necessary to link the scan to a specific " \
                    f"project. \n\nAdditional Information: \n{e}"
            raise SafetyException(message=reason)

        if not status:
            wait_msg = f'Unable to verify "{project_id}". Starting again...'
            time.sleep(on_error_delay)

    return status

def print_project_info(console: Console, project: ProjectModel):
    """
    Print information about the project.

    Args:
        console (Console): The console for output.
        project (ProjectModel): The project model.
    """
    config_msg = "loaded without policies or custom configuration."

    if project.policy:
        if project.policy.source is PolicySource.local:
            rel_location = project.policy.location.name if project.policy.location else ""
            config_msg = "configuration and policies fetched " \
                f"from {rel_location}."
        else:
            config_msg = " policies fetched " \
                "from Safety Platform."

    msg = f"[bold]{project.id} project found[/bold] - {config_msg}"
    console.print(msg)

def print_wait_policy_download(console: Console, closure: Tuple[Any, Dict[str, Any]]) -> Optional[PolicyFileModel]:
    """
    Print a waiting message while downloading a policy from the cloud.

    Args:
        console (Console): The console for output.
        closure (Tuple[Any, Dict[str, Any]]): The function and its arguments to call.

    Returns:
        Optional[PolicyFileModel]: The downloaded policy file model.
    """
    policy = None
    wait_msg = "Looking for a policy from cloud..."

    with console.status(wait_msg, spinner=DEFAULT_SPINNER):
        try:
            f, kwargs = closure
            policy = f(**kwargs)
        except Exception as e:
            LOG.exception(f'Policy download failed, reason: {e}')
            console.print("Not using cloud policy file.")

        if policy:
            wait_msg = "Policy fetched from Safety Platform."
        else:
            # TODO: Send a log
            pass
    return policy


def prompt_project_id(
        console: Console,
        default_id: str) -> str:
    """
    Prompt the user to set a project ID, on a non-interactive mode it will 
    fallback to the default ID parameter.
    """
    default_prj_id = clean_project_id(default_id)

    interactive_mode = console.is_interactive and not console.quiet

    if not interactive_mode:
        LOG.info("Fallback to default project id, because of " \
                 "non-interactive mode.")
        
        return default_prj_id

    hint = f" If empty Safety will use [bold]{default_prj_id}[/bold]"
    prompt_text = f"Set a project id (no spaces).{hint}"

    while True:
        result = Prompt.ask(
            prompt_text,
            console=console,
            default=default_prj_id,
            show_default=False
        )

        return clean_project_id(result) if result != default_prj_id \
            else default_prj_id


def prompt_link_project(console: Console, prj_name: str, prj_admin_email: str) -> bool:
    """
    Prompt the user to link the scan with an existing project.

    Args:
        console (Console): The console for output.
        prj_name (str): The project name.
        prj_admin_email (str): The project admin email.

    Returns:
        bool: True if the user wants to link the scan, False otherwise.
    """
    console.print("[bold]Safety found an existing project with this name in your organization:[/bold]")

    for detail in (f"[bold]Project name:[/bold] {prj_name}",
                   f"[bold]Project admin:[/bold] {prj_admin_email}"):
        console.print(Padding(detail, (0, 0, 0, 2)), emoji=True)

    prompt_question = "Do you want to link it with this existing project?"

    answer = Prompt.ask(prompt=prompt_question, choices=["y", "n"],
                        default="y", show_default=True, console=console).lower()

    return answer == "y"


def render_to_console(cls: Vulnerability, console: Console, rich_kwargs: Dict[str, Any], detailed_output: bool = False) -> None:
    """
    Render a vulnerability to the console.

    Args:
        cls (Vulnerability): The vulnerability instance.
        console (Console): The console for output.
        rich_kwargs (Dict[str, Any]): Additional arguments for rendering.
        detailed_output (bool): Indicates if detailed output is enabled.
    """
    cls.__render__(console, detailed_output, rich_kwargs)


def get_render_console(entity_type: Any) -> Any:
    """
    Get the render function for a specific entity type.

    Args:
        entity_type (Any): The entity type.

    Returns:
        Any: The render function.
    """

    if entity_type is Vulnerability:
        def __render__(self, console: Console, detailed_output: bool, rich_kwargs):
            if not rich_kwargs:
                rich_kwargs = {}

            pre = " Ignored:" if self.ignored else ""
            severity_detail = None

            if self.severity and self.severity.source:
                severity_detail = self.severity.source

                if self.severity.cvssv3 and "base_severity" in self.severity.cvssv3:
                    severity_detail += f", CVSS Severity {self.severity.cvssv3['base_severity'].upper()}"

            advisory_length = 200 if detailed_output else 110

            console.print(
                Padding(
                    f"->{pre} Vuln ID [vuln_id]{self.vulnerability_id}[/vuln_id]: {severity_detail if severity_detail else ''}",
                    (0, 0, 0, 2)
                ), **rich_kwargs)
            console.print(
                Padding(
                    f"{self.advisory[:advisory_length]}{'...' if len(self.advisory) > advisory_length else ''}",
                    (0, 0, 0, 5)
                ), **rich_kwargs)

            if detailed_output:
                console.print(
                    Padding(f"For more information: [link]{self.more_info_url}[/link]", (0, 0, 0, 5)),
                      **rich_kwargs)

        return __render__


def render_scan_html(report: ReportModel, obj: Any) -> str:
    """
    Render the scan report to HTML.

    Args:
        report (ReportModel): The scan report model.
        obj (Any): The object containing additional settings.

    Returns:
        str: The rendered HTML report.
    """
    from safety.scan.command import ScannableEcosystems

    project = report.projects[0] if any(report.projects) else None

    scanned_packages = 0
    affected_packages = 0
    ignored_packages = 0
    remediations_recommended = 0
    ignored_vulnerabilities = 0
    vulnerabilities = 0
    vulns_per_file = defaultdict(int)
    remed_per_file = defaultdict(int)

    for file in project.files:
        scanned_packages += len(file.results.dependencies)
        affected_packages += len(file.results.get_affected_dependencies())
        ignored_vulnerabilities += len(file.results.ignored_vulns)

        for spec in file.results.get_affected_specifications():
            vulnerabilities += len(spec.vulnerabilities)
            vulns_per_file[file.location] += len(spec.vulnerabilities)
            if spec.remediation:
                remed_per_file[file.location] += 1
                remediations_recommended += 1

        ignored_packages += len(file.results.ignored_vulns)

    # TODO: Get this information for the report model (?)
    summary = {"scanned_packages": scanned_packages,
               "affected_packages": affected_packages,
               "remediations_recommended": remediations_recommended,
               "ignored_vulnerabilities": ignored_vulnerabilities, "vulnerabilities": vulnerabilities}

    vulnerabilities = []


    # TODO: This should be based on the configs per command
    ecosystems = [(f"{ecosystem.name.title()}",
                  [file_type.human_name(plural=True) for file_type in ecosystem.file_types]) for ecosystem in [Ecosystem(member.value) for member in list(ScannableEcosystems)]]

    settings ={"audit_and_monitor": True, "platform_url": SAFETY_PLATFORM_URL, "ecosystems": ecosystems}
    template_context = {"report": report, "summary": summary, "announcements": [],
                        "project": project,
                        "platform_enabled": obj.platform_enabled,
                        "settings": settings,
                        "vulns_per_file": vulns_per_file,
                        "remed_per_file": remed_per_file}

    return parse_html(kwargs=template_context, template="scan/index.html")


def generate_spdx_creation_info(spdx_version: str, project_identifier: str) -> Any:
    """
    Generate SPDX creation information.

    Args:
        spdx_version (str): The SPDX version.
        project_identifier (str): The project identifier.

    Returns:
        Any: The SPDX creation information.
    """
    from spdx_tools.spdx.model import (
        Actor,
        ActorType,
        CreationInfo,
    )

    version = int(time.time())
    SPDX_ID_TYPE = "SPDXRef-DOCUMENT"
    DOC_NAME = f"{project_identifier}-{version}"

    DOC_NAMESPACE = f"https://spdx.org/spdxdocs/{DOC_NAME}"
    # DOC_NAMESPACE = f"urn:safety:{project_identifier}:{version}"

    DOC_COMMENT = f"This document was created using SPDX {spdx_version}"
    CREATOR_COMMENT = "Safety CLI automatically created this SPDX document from a scan report."

    TOOL_ID = "safety"
    TOOL_VERSION = get_version()

    doc_creator = Actor(
        actor_type=ActorType.TOOL,
        name=f"{TOOL_ID}-{TOOL_VERSION}",
        email=None
    )

    creation_info = CreationInfo(
        spdx_version=f"SPDX-{spdx_version}",
        spdx_id=SPDX_ID_TYPE,
        name=DOC_NAME,
        document_namespace=DOC_NAMESPACE,
        creators=[doc_creator],
        created=datetime.datetime.now(),
        document_comment=DOC_COMMENT,
        creator_comment=CREATOR_COMMENT
    )
    return creation_info


def create_pkg_ext_ref(*, package: PythonDependency, version: Optional[str]) -> Any:
    """
    Create an external package reference for SPDX.

    Args:
        package (PythonDependency): The package dependency.
        version (Optional[str]): The package version.

    Returns:
        Any: The external package reference.
    """
    from spdx_tools.spdx.model import (
        ExternalPackageRef,
        ExternalPackageRefCategory,
    )

    version_detail = f'@{version}' if version else ''
    pkg_ref = ExternalPackageRef(
        ExternalPackageRefCategory.PACKAGE_MANAGER,
        "purl",
        f"pkg:pypi/{package.name}{version_detail}",
    )
    return pkg_ref


def create_packages(dependencies: List[PythonDependency]) -> List[Any]:
    """
    Create a list of SPDX packages.

    Args:
        dependencies (List[PythonDependency]): List of Python dependencies.

    Returns:
        List[Any]: List of SPDX packages.
    """
    from spdx_tools.spdx.model import (
        Package,
    )
    from spdx_tools.spdx.model.spdx_no_assertion import SpdxNoAssertion

    doc_pkgs = []
    pkgs_added = set([])
    for dependency in dependencies:
        for spec in dependency.specifications:
            pkg_version = next(iter(spec.specifier)).version if spec.is_pinned() else f"{spec.specifier}"
            dep_name = dependency.name.replace('_', '-')
            pkg_id = f"SPDXRef-pip-{dep_name}-{pkg_version}" if spec.is_pinned() else f"SPDXRef-pip-{dep_name}"
            if pkg_id in pkgs_added:
                continue
            pkg_ref = create_pkg_ext_ref(package=dependency, version=pkg_version)

            pkg = Package(
                spdx_id=pkg_id,
                name=f"pip:{dep_name}",
                download_location=SpdxNoAssertion(),
                version=pkg_version,
                file_name="",
                supplier=SpdxNoAssertion(),
                files_analyzed=False,
                license_concluded=SpdxNoAssertion(),
                license_declared=SpdxNoAssertion(),
                copyright_text=SpdxNoAssertion(),
                external_references=[pkg_ref],
            )
            pkgs_added.add(pkg_id)
            doc_pkgs.append(pkg)
    return doc_pkgs


def create_spdx_document(*, report: ReportModel, spdx_version: str) -> Optional[Any]:
    """
    Create an SPDX document.

    Args:
        report (ReportModel): The scan report model.
        spdx_version (str): The SPDX version.

    Returns:
        Optional[Any]: The SPDX document.
    """
    from spdx_tools.spdx.model import (
        Document,
        Relationship,
        RelationshipType,
    )

    project =  report.projects[0] if any(report.projects) else None

    if not project:
        return None

    prj_id = project.id

    if not prj_id:
        parent_name = project.project_path.parent.name
        prj_id = parent_name if parent_name else str(int(time.time()))

    creation_info = generate_spdx_creation_info(spdx_version=spdx_version, project_identifier=prj_id)

    depedencies = iter([])
    for file in project.files:
        depedencies = itertools.chain(depedencies, file.results.dependencies)

    packages = create_packages(depedencies)

    # Requirement for document to have atleast one relationship
    relationship = Relationship(
        "SPDXRef-DOCUMENT",
        RelationshipType.DESCRIBES,
        "SPDXRef-DOCUMENT"
    )
    spdx_doc = Document(
        creation_info,
        packages,
        [],
        [],
        [],
        [relationship],
        []
    )
    return spdx_doc


def render_scan_spdx(report: ReportModel, obj: Any, spdx_version: Optional[str]) -> Optional[Any]:
    """
    Render the scan report to SPDX format.

    Args:
        report (ReportModel): The scan report model.
        obj (Any): The object containing additional settings.
        spdx_version (Optional[str]): The SPDX version.

    Returns:
        Optional[Any]: The rendered SPDX document in JSON format.
    """
    from spdx_tools.spdx.writer.write_utils import convert, validate_and_deduplicate

    # Set to latest supported if a version is not specified
    if not spdx_version:
        spdx_version = "2.3"

    document_obj = create_spdx_document(report=report, spdx_version=spdx_version)
    document_obj = validate_and_deduplicate(document=document_obj, validate=True, drop_duplicates=True)
    doc = None

    if document_obj:
        doc = convert(document=document_obj, converter=None)

    return json.dumps(doc) if doc else None



import http.server
import json
import logging
import random
import socket
import sys
import time
from typing import Any, Optional, Dict, Tuple
import urllib.parse
import threading
import click
from safety.auth.cli_utils import load_auth_session

from safety.auth.utils import is_jupyter_notebook
from safety.console import main_console as console

from safety.auth.constants import AUTH_SERVER_URL, CLI_AUTH_SUCCESS, CLI_LOGOUT_SUCCESS, HOST
from safety.auth.main import save_auth_config
from rich.prompt import Prompt

LOG = logging.getLogger(__name__)


def find_available_port() -> Optional[int]:
    """
    Find an available port on localhost within the dynamic port range (49152-65536).

    Returns:
        Optional[int]: An available port number, or None if no ports are available.
    """
    # Dynamic ports IANA
    port_range = list(range(49152, 65536))
    random.shuffle(port_range)

    for port in port_range:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                result = s.connect(('localhost', port))
                # If the connect succeeds, the port is already in use
            except socket.error as e:
                # If the connect fails, the port is available
                return port

    return None

def auth_process(code: str, state: str, initial_state: str, code_verifier: str, client: Any) -> Any:
    """
    Process the authentication callback and exchange the authorization code for tokens.

    Args:
        code (str): The authorization code.
        state (str): The state parameter from the callback.
        initial_state (str): The initial state parameter.
        code_verifier (str): The code verifier for PKCE.
        client (Any): The OAuth client.

    Returns:
        Any: The user information.

    Raises:
        SystemExit: If there is an error during authentication.
    """
    err = None

    if initial_state is None or initial_state != state:
        err = "The state parameter value provided does not match the expected " \
                "value. The state parameter is used to protect against Cross-Site " \
                "Request Forgery (CSRF) attacks. For security reasons, the " \
                "authorization process cannot proceed with an invalid state " \
                "parameter value. Please try again, ensuring that the state " \
                "parameter value provided in the authorization request matches " \
                "the value returned in the callback."

    if err:
        click.secho(f'Error: {err}', fg='red')
        sys.exit(1)

    try:
        tokens = client.fetch_token(url=f'{AUTH_SERVER_URL}/oauth/token',
                                        code_verifier=code_verifier,
                                        client_id=client.client_id,
                                        grant_type='authorization_code', code=code)

        save_auth_config(access_token=tokens['access_token'],
                            id_token=tokens['id_token'],
                            refresh_token=tokens['refresh_token'])
        return client.fetch_user_info()

    except Exception as e:
        LOG.exception(e)
        sys.exit(1)

class CallbackHandler(http.server.BaseHTTPRequestHandler):
    def auth(self, code: str, state: str, err: str, error_description: str) -> None:
        """
        Handle the authentication callback.

        Args:
            code (str): The authorization code.
            state (str): The state parameter.
            err (str): The error message, if any.
            error_description (str): The error description, if any.
        """
        initial_state = self.server.initial_state
        ctx = self.server.ctx

        result = auth_process(code=code,
                              state=state,
                              initial_state=initial_state,
                              code_verifier=ctx.obj.auth.code_verifier,
                              client=ctx.obj.auth.client)

        self.server.callback = result
        self.do_redirect(location=CLI_AUTH_SUCCESS, params={})

    def logout(self) -> None:
        """
        Handle the logout callback.
        """
        ctx = self.server.ctx
        uri = CLI_LOGOUT_SUCCESS

        if ctx.obj.auth.org:
            uri = f"{uri}&org_id={ctx.obj.auth.org.id}"

        self.do_redirect(location=CLI_LOGOUT_SUCCESS, params={})

    def do_GET(self) -> None:
        """
        Handle GET requests.
        """
        query = urllib.parse.urlparse(self.path).query
        params = urllib.parse.parse_qs(query)
        callback_type: Optional[str] = None

        try:
            c_type = params.get('type', [])
            if isinstance(c_type, list) and len(c_type) == 1 and isinstance(c_type[0], str):
                callback_type = c_type[0]
        except Exception:
            msg = "Unable to process the callback, try again."
            self.send_error(400, msg)
            click.secho("Unable to process the callback, try again.")
            return

        if callback_type == 'logout':
            self.logout()
            return

        code = params.get('code', [''])[0]
        state = params.get('state', [''])[0]
        err = params.get('error', [''])[0]
        error_description = params.get('error_description', [''])[0]

        self.auth(code=code, state=state, err=err, error_description=error_description)

    def do_redirect(self, location: str, params: Dict) -> None:
        """
        Redirect the client to the specified location.

        Args:
            location (str): The URL to redirect to.
            params (dict): Additional parameters for the redirection.
        """
        self.send_response(302)
        self.send_header('Location', location)
        self.send_header('Connection', 'close')
        self.send_header('Cache-Control', 'no-store, no-cache, must-revalidate')
        self.end_headers()

    def log_message(self, format: str, *args: Any) -> None:
        """
        Log an arbitrary message.

        Args:
            format (str): The format string.
            args (Any): Arguments for the format string.
        """
        LOG.info(format % args)


def process_browser_callback(uri: str, **kwargs: Any) -> Any:
    """
    Process the browser callback for authentication.

    Args:
        uri (str): The authorization URL.
        **kwargs (Any): Additional keyword arguments.

    Returns:
        Any: The user information.

    Raises:
        SystemExit: If there is an error during the process.
    """

    class ThreadedHTTPServer(http.server.HTTPServer):
        def __init__(self, server_address: Tuple, RequestHandlerClass: Any) -> None:
            """
            Initialize the ThreadedHTTPServer.
            Args:
                server_address (Tuple): The server address as a tuple (host, port).
                RequestHandlerClass (Any): The request handler class.
            """
            super().__init__(server_address, RequestHandlerClass)
            self.initial_state = None
            self.ctx = None
            self.callback = None
            self.timeout_reached = False

        def handle_timeout(self) -> None:
            """
            Handle server timeout.
            """
            self.timeout_reached = True
            return super().handle_timeout()

    PORT = find_available_port()

    if not PORT:
        click.secho("No available ports.")
        sys.exit(1)

    try:
        headless = kwargs.get("headless", False)
        initial_state = kwargs.get("initial_state", None)
        ctx = kwargs.get("ctx", None)
        message = "Copy and paste this URL into your browser:\n⚠️ Ensure there are no extra spaces, especially at line breaks, as they may break the link."

        if not headless:
            # Start a threaded HTTP server to handle the callback
            server = ThreadedHTTPServer((HOST, PORT), CallbackHandler)
            server.initial_state = initial_state
            server.timeout = kwargs.get("timeout", 600)
            server.ctx = ctx
            server_thread = threading.Thread(target=server.handle_request)
            server_thread.start()
            message = f"If the browser does not automatically open in 5 seconds, copy and paste this url into your browser:"

        target = uri if headless else f"{uri}&port={PORT}"

        if is_jupyter_notebook():
            console.print(f"{message} {target}")
        else:
            console.print(f"{message} [link={target}]{target}[/link]")
        
        if headless:
            # Handle the headless mode where user manually provides the response
            exchange_data = None
            while not exchange_data:
                auth_code_text = Prompt.ask("Paste the response here", default=None, console=console)
                try:
                    exchange_data = json.loads(auth_code_text)
                    state = exchange_data["state"]
                    code = exchange_data["code"]
                except Exception:
                    code = state = None

            return auth_process(code=code,
                                state=state,
                                initial_state=initial_state,
                                code_verifier=ctx.obj.auth.code_verifier,
                                client=ctx.obj.auth.client)
        else:
            # Wait for the browser authentication in non-headless mode
            wait_msg = "waiting for browser authentication"
            with console.status(wait_msg, spinner="bouncingBar"):
                time.sleep(2)
                click.launch(target)
                server_thread.join()

    except OSError as e:
        if e.errno == socket.errno.EADDRINUSE:
            reason = f"The port {HOST}:{PORT} is currently being used by another application or process. Please choose a different port or terminate the conflicting application/process to free up the port."
        else:
            reason = "An error occurred while performing this operation."

        click.secho(reason)
        sys.exit(1)

    return server.callback



from __future__ import unicode_literals

from packaging.version import parse as parse_version
from packaging.specifiers import SpecifierSet
import requests
from typing import Any, Optional, Generator, Tuple, List

from datetime import datetime
from dparse import parse, parser, updater, filetypes
from dparse.dependencies import Dependency
from dparse.parser import setuptools_parse_requirements_backport as parse_requirements


class RequirementFile(object):
    """
    Class representing a requirements file with its content and metadata.

    Attributes:
        path (str): The file path.
        content (str): The content of the file.
        sha (Optional[str]): The SHA of the file.
    """
    def __init__(self, path: str, content: str, sha: Optional[str] = None):
        self.path = path
        self.content = content
        self.sha = sha
        self._requirements = None
        self._other_files = None
        self._is_valid = None
        self.is_pipfile = False
        self.is_pipfile_lock = False
        self.is_setup_cfg = False

    def __str__(self) -> str:
        return "RequirementFile(path='{path}', sha='{sha}', content='{content}')".format(
            path=self.path,
            content=self.content[:30] + "[truncated]" if len(self.content) > 30 else self.content,
            sha=self.sha
        )

    @property
    def is_valid(self) -> Optional[bool]:
        """
        Checks if the requirements file is valid by parsing it.

        Returns:
            bool: True if the file is valid, False otherwise.
        """
        if self._is_valid is None:
            self._parse()
        return self._is_valid

    @property
    def requirements(self) -> Optional[List]:
        """
        Returns the list of requirements parsed from the file.

        Returns:
            List: The list of requirements.
        """
        if not self._requirements:
            self._parse()
        return self._requirements

    @property
    def other_files(self) -> Optional[List]:
        """
        Returns the list of other files resolved from the requirements file.

        Returns:
            List: The list of other files.
        """
        if not self._other_files:
            self._parse()
        return self._other_files

    @staticmethod
    def parse_index_server(line: str) -> Optional[str]:
        """
        Parses the index server from a given line.

        Args:
            line (str): The line to parse.

        Returns:
            str: The parsed index server.
        """
        return parser.Parser.parse_index_server(line)

    def _hash_parser(self, line: str) -> Optional[Tuple[str, List[str]]]:
        """
        Parses the hashes from a given line.

        Args:
            line (str): The line to parse.

        Returns:
            List: The list of parsed hashes.
        """
        return parser.Parser.parse_hashes(line)

    def _parse_requirements_txt(self) -> None:
        """
        Parses the requirements.txt file format.
        """
        self.parse_dependencies(filetypes.requirements_txt)

    def _parse_conda_yml(self) -> None:
        """
        Parses the conda.yml file format.
        """
        self.parse_dependencies(filetypes.conda_yml)

    def _parse_tox_ini(self) -> None:
        """
        Parses the tox.ini file format.
        """
        self.parse_dependencies(filetypes.tox_ini)

    def _parse_pipfile(self) -> None:
        """
        Parses the Pipfile format.
        """
        self.parse_dependencies(filetypes.pipfile)
        self.is_pipfile = True

    def _parse_pipfile_lock(self) -> None:
        """
        Parses the Pipfile.lock format.
        """
        self.parse_dependencies(filetypes.pipfile_lock)
        self.is_pipfile_lock = True

    def _parse_setup_cfg(self) -> None:
        """
        Parses the setup.cfg format.
        """
        self.parse_dependencies(filetypes.setup_cfg)
        self.is_setup_cfg = True

    def _parse(self) -> None:
        """
        Parses the requirements file to extract dependencies and other files.
        """
        self._requirements, self._other_files = [], []
        if self.path.endswith('.yml') or self.path.endswith(".yaml"):
            self._parse_conda_yml()
        elif self.path.endswith('.ini'):
            self._parse_tox_ini()
        elif self.path.endswith("Pipfile"):
            self._parse_pipfile()
        elif self.path.endswith("Pipfile.lock"):
            self._parse_pipfile_lock()
        elif self.path.endswith('setup.cfg'):
            self._parse_setup_cfg()
        else:
            self._parse_requirements_txt()
        self._is_valid = len(self._requirements) > 0 or len(self._other_files) > 0

    def parse_dependencies(self, file_type: str) -> None:
        """
        Parses the dependencies from the content based on the file type.

        Args:
            file_type (str): The type of the file.
        """
        result = parse(
            self.content,
            path=self.path,
            sha=self.sha,
            file_type=file_type,
            marker=(
                ("pyup: ignore file", "pyup:ignore file"),  # file marker
                ("pyup: ignore", "pyup:ignore"),  # line marker
            )
        )
        for dep in result.dependencies:
            req = Requirement(
                name=dep.name,
                specs=dep.specs,
                line=dep.line,
                lineno=dep.line_numbers[0] if dep.line_numbers else 0,
                extras=dep.extras,
                file_type=file_type,
            )
            req.index_server = dep.index_server
            if self.is_pipfile:
                req.pipfile = self.path
            req.hashes = dep.hashes
            self._requirements.append(req)
        self._other_files = result.resolved_files

    def iter_lines(self, lineno: int = 0) -> Generator[str, None, None]:
        """
        Iterates over lines in the content starting from a specific line number.

        Args:
            lineno (int): The line number to start from.

        Yields:
            str: The next line in the content.
        """
        for line in self.content.splitlines()[lineno:]:
            yield line

    @classmethod
    def resolve_file(cls, file_path: str, line: str) -> str:
        """
        Resolves a file path from a given line.

        Args:
            file_path (str): The file path to resolve.
            line (str): The line containing the file path.

        Returns:
            str: The resolved file path.
        """
        return parser.Parser.resolve_file(file_path, line)


class Requirement(object):
    """
    Class representing a single requirement.

    Attributes:
        name (str): The name of the requirement.
        specs (SpecifierSet): The version specifiers for the requirement.
        line (str): The line containing the requirement.
        lineno (int): The line number of the requirement.
        extras (List): The extras for the requirement.
        file_type (str): The type of the file containing the requirement.
    """
    def __init__(self, name: str, specs: SpecifierSet, line: str, lineno: int, extras: List, file_type: str):
        self.name = name
        self.key = name.lower()
        self.specs = specs
        self.line = line
        self.lineno = lineno
        self.index_server = None
        self.extras = extras
        self.hashes = []
        self.file_type = file_type
        self.pipfile = None

        self.hashCmp = (
            self.key,
            self.specs,
            frozenset(self.extras),
        )

        self._is_insecure = None
        self._changelog = None

        # Convert compatible releases to a range of versions
        if len(self.specs._specs) == 1 and next(iter(self.specs._specs))._spec[0] == "~=":
            # convert compatible releases to something more easily consumed,
            # e.g. '~=1.2.3' is equivalent to '>=1.2.3,<1.3.0', while '~=1.2'
            # is equivalent to '>=1.2,<2.0'
            min_version = next(iter(self.specs._specs))._spec[1]
            max_version = list(parse_version(min_version).release)
            max_version[-1] = 0
            max_version[-2] = max_version[-2] + 1
            max_version = '.'.join(str(x) for x in max_version)

            self.specs = SpecifierSet('>=%s,<%s' % (min_version, max_version))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Requirement) and
            self.hashCmp == other.hashCmp
        )

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __str__(self) -> str:
        return "Requirement.parse({line}, {lineno})".format(line=self.line, lineno=self.lineno)

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def is_pinned(self) -> bool:
        """
        Checks if the requirement is pinned to a specific version.

        Returns:
            bool: True if pinned, False otherwise.
        """
        if len(self.specs._specs) == 1 and next(iter(self.specs._specs))._spec[0] == "==":
            return True
        return False

    @property
    def is_open_ranged(self) -> bool:
        """
        Checks if the requirement has an open range of versions.

        Returns:
            bool: True if open ranged, False otherwise.
        """
        if len(self.specs._specs) == 1 and next(iter(self.specs._specs))._spec[0] == ">=":
            return True
        return False

    @property
    def is_ranged(self) -> bool:
        """
        Checks if the requirement has a range of versions.

        Returns:
            bool: True if ranged, False otherwise.
        """
        return len(self.specs._specs) >= 1 and not self.is_pinned

    @property
    def is_loose(self) -> bool:
        """
        Checks if the requirement has no version specifiers.

        Returns:
            bool: True if loose, False otherwise.
        """
        return len(self.specs._specs) == 0

    @staticmethod
    def convert_semver(version: str) -> dict:
        """
        Converts a version string to a semantic version dictionary.

        Args:
            version (str): The version string.

        Returns:
            dict: The semantic version dictionary.
        """
        semver = {'major': 0, "minor": 0, "patch": 0}
        version = version.split(".")
        # don't be overly clever here. repitition makes it more readable and works exactly how
        # it is supposed to
        try:
            semver['major'] = int(version[0])
            semver['minor'] = int(version[1])
            semver['patch'] = int(version[2])
        except (IndexError, ValueError):
            pass
        return semver

    @property
    def can_update_semver(self) -> bool:
        """
        Checks if the requirement can be updated based on semantic versioning rules.

        Returns:
            bool: True if it can be updated, False otherwise.
        """
        # return early if there's no update filter set
        if "pyup: update" not in self.line:
            return True
        update = self.line.split("pyup: update")[1].strip().split("#")[0]
        current_version = Requirement.convert_semver(next(iter(self.specs._specs))._spec[1])
        next_version = Requirement.convert_semver(self.latest_version)
        if update == "major":
            if current_version['major'] < next_version['major']:
                return True
        elif update == 'minor':
            if current_version['major'] < next_version['major'] \
                    or current_version['minor'] < next_version['minor']:
                return True
        return False

    @property
    def filter(self):
        """
        Returns the filter for the requirement if specified.

        Returns:
            Optional[SpecifierSet]: The filter specifier set, or None if not specified.
        """
        rqfilter = False
        if "rq.filter:" in self.line:
            rqfilter = self.line.split("rq.filter:")[1].strip().split("#")[0]
        elif "pyup:" in self.line:
            if "pyup: update" not in self.line:
                rqfilter = self.line.split("pyup:")[1].strip().split("#")[0]
                # unset the filter once the date set in 'until' is reached
                if "until" in rqfilter:
                    rqfilter, until = [l.strip() for l in rqfilter.split("until")]
                    try:
                        until = datetime.strptime(until, "%Y-%m-%d")
                        if until < datetime.now():
                            rqfilter = False
                    except ValueError:
                        # wrong date formatting
                        pass
        if rqfilter:
            try:
                rqfilter, = parse_requirements("filter " + rqfilter)
                if len(rqfilter.specifier._specs) > 0:
                    return rqfilter.specifier
            except ValueError:
                pass
        return False

    @property
    def version(self) -> Optional[str]:
        """
        Returns the current version of the requirement.

        Returns:
            Optional[str]: The current version, or None if not pinned.
        """
        if self.is_pinned:
            return next(iter(self.specs._specs))._spec[1]

        specs = self.specs
        if self.filter:
            specs = SpecifierSet(
                ",".join(["".join(s._spec) for s in list(specs._specs) + list(self.filter._specs)])
            )
        return self.get_latest_version_within_specs(
            specs,
            versions=self.package.versions,
            prereleases=self.prereleases
        )

    def get_hashes(self, version: str) -> List:
        """
        Retrieves the hashes for a specific version from PyPI.

        Args:
            version (str): The version to retrieve hashes for.

        Returns:
            List: A list of hashes for the specified version.
        """
        r = requests.get('https://pypi.org/pypi/{name}/{version}/json'.format(
            name=self.key,
            version=version
        ))
        hashes = []
        data = r.json()

        for item in data.get("urls", {}):
            sha256 = item.get("digests", {}).get("sha256", False)
            if sha256:
                hashes.append({"hash": sha256, "method": "sha256"})
        return hashes

    def update_version(self, content: str, version: str, update_hashes: bool = True) -> str:
        """
        Updates the version of the requirement in the content.

        Args:
            content (str): The original content.
            version (str): The new version to update to.
            update_hashes (bool): Whether to update the hashes as well.

        Returns:
            str: The updated content.
        """
        if self.file_type == filetypes.tox_ini:
            updater_class = updater.ToxINIUpdater
        elif self.file_type == filetypes.conda_yml:
            updater_class = updater.CondaYMLUpdater
        elif self.file_type == filetypes.requirements_txt:
            updater_class = updater.RequirementsTXTUpdater
        elif self.file_type == filetypes.pipfile:
            updater_class = updater.PipfileUpdater
        elif self.file_type == filetypes.pipfile_lock:
            updater_class = updater.PipfileLockUpdater
        elif self.file_type == filetypes.setup_cfg:
            updater_class = updater.SetupCFGUpdater
        else:
            raise NotImplementedError

        dep = Dependency(
            name=self.name,
            specs=self.specs,
            line=self.line,
            line_numbers=[self.lineno, ] if self.lineno != 0 else None,
            dependency_type=self.file_type,
            hashes=self.hashes,
            extras=self.extras
        )
        hashes = []
        if self.hashes and update_hashes:
            hashes = self.get_hashes(version)

        return updater_class.update(
            content=content,
            dependency=dep,
            version=version,
            hashes=hashes,
            spec="=="
        )

    @classmethod
    def parse(cls, s: str, lineno: int, file_type: str = filetypes.requirements_txt) -> 'Requirement':
        """
        Parses a requirement from a line of text.

        Args:
            s (str): The line of text.
            lineno (int): The line number.
            file_type (str): The type of the file containing the requirement.

        Returns:
            Requirement: The parsed requirement.
        """
        # setuptools requires a space before the comment. If this isn't the case, add it.
        if "\t#" in s:
            parsed, = parse_requirements(s.replace("\t#", "\t #"))
        else:
            parsed, = parse_requirements(s)

        return cls(
            name=parsed.name,
            specs=parsed.specifier,
            line=s,
            lineno=lineno,
            extras=parsed.extras,
            file_type=file_type
        )



from pathlib import Path
import unittest
from unittest.mock import Mock, patch
from safety.auth.constants import CLI_AUTH, CLI_AUTH_LOGOUT, CLI_CALLBACK

from safety.auth import main
from safety.auth.main import get_authorization_data, \
    get_logout_url, get_organization, get_redirect_url



class TestAuthMain(unittest.TestCase):

    def setUp(self):
        self.assets = Path(__file__).parent / Path("test_assets/")

    def tearDown(self):
        pass

    def test_get_authorization_data(self):
        org_id = "org_id3dasdasd"
        client = Mock()
        code_verifier = "test_code_verifier"
        organization = Mock(id=org_id)

        client.create_authorization_url = Mock()
        _ = get_authorization_data(client, code_verifier, organization)

        kwargs = {
             "sign_up": False,
             "locale": "en",
             "ensure_auth": False,
             "organization": org_id,
             "headless": False
        }

        client.create_authorization_url.assert_called_once_with(
             CLI_AUTH, code_verifier=code_verifier, **kwargs)
        
        client.create_authorization_url = Mock()
        _ = get_authorization_data(client, code_verifier, organization=None)

        kwargs = {
             "sign_up": False,
             "locale": "en",
             "ensure_auth":False,
             "headless": False             
        }

        client.create_authorization_url.assert_called_once_with(
             CLI_AUTH, code_verifier=code_verifier, **kwargs)        

    def get_logout_url(self, id_token):
        return f'{CLI_AUTH_LOGOUT}?id_token={id_token}'

    def test_get_logout_url(self):
        id_token = "test_id_token"
        result = get_logout_url(id_token)
        expected_result = f'{CLI_AUTH_LOGOUT}?id_token={id_token}'
        self.assertEqual(result, expected_result)

    def test_get_redirect_url(self):
        self.assertEqual(get_redirect_url(), CLI_CALLBACK)

    def test_get_organization(self):
        with patch.object(main, "CONFIG",
                          (self.assets / Path("config.ini")).absolute()):
            result = get_organization()
            self.assertIsNotNone(result)
            self.assertEqual(result.id, "org_id23423ds")
            self.assertEqual(result.name, "Safety CLI Org")
        


from pathlib import Path
from typing import Optional

from pydantic.dataclasses import dataclass


@dataclass
class UnverifiedProjectModel:
    """
    Data class representing an unverified project model.
    """

    id: Optional[str]
    project_path: Path
    created: bool
    name: Optional[str] = None
    url_path: Optional[str] = None
